from qtpy.QtWidgets import (
    QWidget, QFormLayout, QComboBox, QDoubleSpinBox, QSlider, QLabel, QPushButton, QFileDialog
)
from qtpy.QtCore import Qt

import matplotlib.pyplot as plt 
from matplotlib.backends.backend_qt5agg import ( FigureCanvasQTAgg as FigureCanvas )

class OneatVisWidget(QWidget):
    
    def __init__(self, parent = None):
        
        super().__init__(parent = parent)
        
        self._layout = QFormLayout(parent= self)
        
        self.detectionidbox = QPushButton('Select Prediction File (csv)', parent = self)
        
        
        
        self.startprobspinbox = QDoubleSpinBox()
        self.startprobspinbox.setValue(0.9)
        self.startprobspinbox.setDecimals(10)
        
        self.scoreslider = QSlider(Qt.Horizontal, parent = self)
        self.scoreslider.setToolTip('Scroll through probability score')
        self.scoreslider.setRange(0, 5000)
        self.scoreslider.setSingleStep(1)
        self.scoreslider.setTickInterval(1)
        self.scoreslider.setValue(0)
        
        self.label = QLabel(parent= self)
        self.label.setAlignment(Qt.AlignCenter | Qt.AlignVCenter)
        self.label.setMinimumWidth(80)
        self.label.setText(f"{0.9:.5f}")
        
        self.recomputebutton = QPushButton("Recompute with changed parameters",parent= self)
        
        
        self.figure = plt.figure(figsize=(4, 4))
        self.multiplot_widget = FigureCanvas(self.figure)
        self.multiplot_widget.setMinimumSize(200, 200)
        self.ax = self.multiplot_widget.figure.subplots(1, 1)

        self._layout.addWidget(self.multiplot_widget)
        self._layout.addRow(self.detectionidbox)
        self._layout.addRow("Lowest probability event", self.startprobspinbox)
        self._layout.addRow("Veto", self.label)
        self._layout.addRow("Score slider", self.scoreslider)
        self._layout.addRow(self.recomputebutton)
        
        
        
    
    


#!/usr/bin/env python3

import argparse


class TrainConfig(argparse.Namespace):
    def __init__(
        self,
        cell_type_name,
        cell_type_label,
        cell_position_name,
        cell_position_label,
        **kwargs
    ):

        self.cell_type_name = cell_type_name
        self.cell_type_label = cell_type_label
        self.cell_position_name = cell_position_name
        self.cell_position_label = cell_position_label
        assert len(cell_type_name) == len(cell_type_label)
        assert len(cell_position_name) == len(cell_position_label)

    def to_json(self):

        config = {}
        configCord = {}
        for i in range(0, len(self.cell_type_name)):

            config[self.cell_type_name[i]] = self.cell_type_label[i]
        for i in range(0, len(self.cell_position_name)):

            configCord[self.cell_position_name[i]] = self.cell_position_label[
                i
            ]

        return config, configCord



from ._qt import OneatWidget
from .OneatScoreKeeper import ScoreModels
from .OneatVisualization import OneatVisualization
from .OneatVolumeVisualization import OneatVolumeVisualization

__all__ = (
    "OneatWidget",
    "OneatVisualization",
    "OneatVolumeVisualization",
    "ScoreModels",
)



import csv
import os
from pathlib import Path

import numpy as np
import pandas as pd
from napari import Viewer, layers
from scipy import spatial
from scipy.ndimage import zoom
from skimage import measure, morphology
from tifffile import imread, imwrite

from ..utils import location_map


class OneatVisualization:
    def __init__(
        self,
        viewer: Viewer,
        key_categories: dict,
        csvdir: str,
        savedir: str,
        savename: str,
        ax,
        figure,
    ):

        self.viewer = viewer
        self.csvdir = csvdir
        self.savedir = savedir
        self.savename = savename
        self.key_categories = key_categories
        self.ax = ax
        self.figure = figure
        self.dataset = None
        self.event_name = None
        self.cell_count = None
        self.image = None
        self.seg_image = None
        self.event_locations = []
        self.event_locations_dict = {}
        self.event_locations_size_dict = {}
        self.size_locations = []
        self.score_locations = []
        self.confidence_locations = []
        self.event_locations_clean = []
        self.cleantimelist = []
        self.cleaneventlist = []
        self.cleannormeventlist = []
        self.cleancelllist = []
        self.labelsize = {}
        self.segimagedir = None
        self.plot_event_name = None
        self.event_count_plot = None
        self.event_norm_count_plot = None
        self.cell_count_plot = None
        self.imagename = None
        self.originalimage = None

    # To prevent early detectin of events
    def cluster_points(self, nms_space):

        print("before", len(self.event_locations_size_dict))

        for (k, v) in self.event_locations_dict.items():
            currenttime = k
            event_locations = v

            if len(event_locations) > 0:
                tree = spatial.cKDTree(event_locations)
                forwardtime = currenttime + 1
                if int(forwardtime) in self.event_locations_dict.keys():
                    forward_event_locations = self.event_locations_dict[
                        int(forwardtime)
                    ]
                    for location in forward_event_locations:
                        if (
                            int(forwardtime),
                            int(location[0]),
                            int(location[1]),
                        ) in self.event_locations_size_dict:
                            (
                                forwardsize,
                                forwardscore,
                                forwardconfidence,
                            ) = self.event_locations_size_dict[
                                int(forwardtime),
                                int(location[0]),
                                int(location[1]),
                            ]
                            distance, nearest_location = tree.query(location)
                            nearest_location = int(
                                event_locations[nearest_location][0]
                            ), int(event_locations[nearest_location][1])

                            if distance <= nms_space:
                                if (
                                    int(currenttime),
                                    int(nearest_location[0]),
                                    int(nearest_location[1]),
                                ) in self.event_locations_size_dict:
                                    (
                                        currentsize,
                                        currentscore,
                                        currentconfidence,
                                    ) = self.event_locations_size_dict[
                                        int(currenttime),
                                        int(nearest_location[0]),
                                        int(nearest_location[1]),
                                    ]
                                    if currentscore >= forwardscore:
                                        self.event_locations_size_dict.pop(
                                            (
                                                int(forwardtime),
                                                int(location[0]),
                                                int(location[1]),
                                            )
                                        )

                                    if currentscore < forwardscore:
                                        self.event_locations_size_dict.pop(
                                            (
                                                int(currenttime),
                                                int(nearest_location[0]),
                                                int(nearest_location[1]),
                                            )
                                        )

        print("after", len(self.event_locations_size_dict))
        self.show_clean_csv()

    def show_clean_csv(self):
        self.cleaneventlist = []
        self.cleantimelist = []
        self.event_locations_clean.clear()
        dict_locations = self.event_locations_size_dict.keys()
        tlocations = []
        zlocations = []
        ylocations = []
        xlocations = []
        scores = []
        radiuses = []
        confidences = []
        for location, sizescore in self.event_locations_size_dict.items():
            tlocations.append(float(location[0]))
            if len(self.originalimage.shape) == 4:
                zlocations.append(float(self.originalimage.shape[1] // 2))
            else:
                zlocations.append(0)
            ylocations.append(float(location[1]))
            xlocations.append(float(location[2]))

            scores.append(float(sizescore[1]))
            radiuses.append(float(sizescore[0]))
            confidences.append(float(sizescore[2]))
        for location in dict_locations:
            self.event_locations_clean.append(location)

        event_count = np.column_stack(
            [
                tlocations,
                zlocations,
                ylocations,
                xlocations,
                scores,
                radiuses,
                confidences,
            ]
        )
        event_count = sorted(event_count, key=lambda x: x[0], reverse=False)

        event_data = []
        csvname = self.savedir + "/" + "Clean" + self.event_name + "Location"
        if os.path.exists(csvname + ".csv"):
            os.remove(csvname + ".csv")
        writer = csv.writer(open(csvname + ".csv", "a", newline=""))
        filesize = os.stat(csvname + ".csv").st_size

        if filesize < 1:
            writer.writerow(
                ["T", "Z", "Y", "X", "Score", "Size", "Confidence"]
            )
        for line in event_count:
            if line not in event_data:
                event_data.append(line)
            writer.writerows(event_data)
            event_data = []
        name_remove = ("Clean Detections", "Clean Location Map")

        point_properties = {
            "score": scores,
            "confidence": confidences,
            "size": radiuses,
        }

        for layer in list(self.viewer.layers):

            if any(name in layer.name for name in name_remove):
                self.viewer.layers.remove(layer)
        self.viewer.add_points(
            self.event_locations_clean,
            properties=point_properties,
            name="Clean Detections",
            face_color=[0] * 4,
            edge_color="green",
        )

        df = pd.DataFrame(self.event_locations_clean, columns=["T", "Y", "X"])
        T_pred = df[df.keys()[0]][0:]
        listtime_pred = T_pred.tolist()

        for j in range(self.image.shape[0]):
            cleanlist = []
            for i in range(len(listtime_pred)):

                if j == listtime_pred[i]:
                    cleanlist.append(listtime_pred[i])

            countT = len(cleanlist)
            self.cleantimelist.append(j)
            self.cleaneventlist.append(countT)

    def show_plot(
        self,
        plot_event_name,
        event_count_plot,
        event_norm_count_plot,
        cell_count_plot,
        segimagedir=None,
        event_threshold=0,
    ):

        timelist = []
        eventlist = []
        normeventlist = []
        celllist = []
        self.ax.cla()

        self.segimagedir = segimagedir
        self.plot_event_name = plot_event_name
        self.event_count_plot = event_count_plot
        self.event_norm_count_plot = event_norm_count_plot
        self.cell_count_plot = cell_count_plot

        if self.dataset is not None:

            for layer in list(self.viewer.layers):
                if isinstance(layer, layers.Image):
                    self.image = layer.data
                if isinstance(layer, layers.Labels):
                    self.seg_image = layer.data

            if self.image is not None:
                currentT = np.round(self.dataset["T"]).astype("int")
                currentsize = self.dataset["Score"]

                for i in range(0, self.image.shape[0]):

                    condition = currentT == i
                    condition_indices = self.dataset_index[condition]
                    conditionScore = currentsize[condition_indices]
                    score_condition = conditionScore > event_threshold
                    countT = len(conditionScore[score_condition])
                    timelist.append(i)
                    eventlist.append(countT)
                    if (
                        self.segimagedir is not None
                        and self.seg_image is not None
                    ):

                        all_cells = self.cell_count[i]
                        celllist.append(all_cells + 1)
                        normeventlist.append(countT / (all_cells + 1))
                self.cleannormeventlist = []
                if len(self.cleaneventlist) > 0:
                    for k in range(len(self.cleaneventlist)):
                        self.cleannormeventlist.append(
                            self.cleaneventlist[k] / celllist[k]
                        )
                if self.plot_event_name == self.event_count_plot:
                    self.ax.plot(timelist, eventlist, "-r")
                    self.ax.plot(self.cleantimelist, self.cleaneventlist, "-g")
                    self.ax.set_title(self.event_name + "Events")
                    self.ax.set_xlabel("Time")
                    self.ax.set_ylabel("Counts")
                    self.figure.canvas.draw()
                    self.figure.canvas.flush_events()

                    self.figure.savefig(
                        self.savedir
                        + self.event_name
                        + self.event_count_plot
                        + ".png",
                        dpi=300,
                    )

                if (
                    self.plot_event_name == self.event_norm_count_plot
                    and len(normeventlist) > 0
                ):
                    self.ax.plot(timelist, normeventlist, "-r")
                    self.ax.plot(
                        self.cleantimelist, self.cleannormeventlist, "-g"
                    )
                    self.ax.set_title(self.event_name + "Normalized Events")
                    self.ax.set_xlabel("Time")
                    self.ax.set_ylabel("Normalized Counts")
                    self.figure.canvas.draw()
                    self.figure.canvas.flush_events()

                    self.figure.savefig(
                        self.savedir
                        + self.event_name
                        + self.event_norm_count_plot
                        + ".png",
                        dpi=300,
                    )

                if (
                    self.plot_event_name == self.cell_count_plot
                    and len(celllist) > 0
                ):
                    self.ax.plot(timelist, celllist, "-r")
                    self.ax.set_title("Total Cell counts")
                    self.ax.set_xlabel("Time")
                    self.ax.set_ylabel("Total Cell Counts")
                    self.figure.canvas.draw()
                    self.figure.canvas.flush_events()
                    self.figure.savefig(
                        self.savedir + self.cell_count_plot + ".png", dpi=300
                    )

    def show_image(
        self,
        image_toread,
        imagename,
        segimagedir=None,
        heatmapimagedir=None,
        heatname="_Heat",
        start_project_mid=0,
        end_project_mid=0,
    ):
        self.imagename = imagename
        name_remove = ("Image", "SegImage")
        for layer in list(self.viewer.layers):
            if any(name in layer.name for name in name_remove):
                self.viewer.layers.remove(layer)
        try:
            self.image = imread(image_toread)

            if heatmapimagedir is not None:
                try:
                    heat_image = imread(
                        heatmapimagedir + imagename + heatname + ".tif"
                    )
                except FileNotFoundError:
                    heat_image = None

            if segimagedir is not None:
                self.seg_image = imread(segimagedir + imagename + ".tif")

                if (
                    start_project_mid is not None
                    or end_project_mid is not None
                ):
                    if len(self.seg_image.shape) == 4:
                        self.seg_image = MidSlices(
                            self.seg_image,
                            start_project_mid,
                            end_project_mid,
                            axis=1,
                        )

                self.viewer.add_labels(
                    self.seg_image.astype("uint16"),
                    name="SegImage" + imagename,
                )

            if len(self.image.shape) == 4:
                self.originalimage = self.image
                if (
                    start_project_mid is not None
                    or end_project_mid is not None
                ):
                    self.image = MidSlices(
                        self.image, start_project_mid, end_project_mid, axis=1
                    )

            else:
                self.originalimage = self.image
            self.viewer.add_image(self.image, name="Image" + imagename)
            if heatmapimagedir is not None:
                try:
                    self.viewer.add_image(
                        heat_image,
                        name="Image" + imagename + heatname,
                        blending="additive",
                        colormap="inferno",
                    )
                except FileNotFoundError:
                    pass

        except FileNotFoundError:
            pass

    def show_csv(
        self,
        imagename,
        csv_event_name,
        segimagedir=None,
        event_threshold=0,
        heatmapsteps=0,
        nms_space=0,
    ):

        csvname = None
        self.event_locations_size_dict.clear()
        self.size_locations = []
        self.score_locations = []
        self.event_locations = []
        self.confidence_locations = []
        self.ax.cla()
        for layer in list(self.viewer.layers):
            if "Detections" in layer.name or layer.name in "Detections":
                self.viewer.layers.remove(layer)
        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0 and csv_event_name == event_name:
                self.event_label = event_label
                csvname = list(Path(self.csvdir).glob("*.csv"))[0]
        if csvname is not None:

            self.event_name = csv_event_name
            self.dataset = pd.read_csv(csvname, delimiter=",")
            nrows = len(self.dataset.columns)
            for index, row in self.dataset.iterrows():
                tcenter = int(row[0])
                ycenter = row[2]
                xcenter = row[3]
                if nrows > 4:
                    score = row[4]
                    size = row[5]
                    confidence = row[6]
                else:
                    score = 1.0
                    size = 10
                    confidence = 1.0
                self.dataset_index = self.dataset.index
                if score > event_threshold:
                    self.event_locations.append(
                        [int(tcenter), int(ycenter), int(xcenter)]
                    )

                    if int(tcenter) in self.event_locations_dict.keys():
                        current_list = self.event_locations_dict[int(tcenter)]
                        current_list.append([int(ycenter), int(xcenter)])
                        self.event_locations_dict[int(tcenter)] = current_list
                        self.event_locations_size_dict[
                            (int(tcenter), int(ycenter), int(xcenter))
                        ] = [size, score, confidence]
                    else:
                        current_list = []
                        current_list.append([int(ycenter), int(xcenter)])
                        self.event_locations_dict[int(tcenter)] = current_list
                        self.event_locations_size_dict[
                            int(tcenter), int(ycenter), int(xcenter)
                        ] = [size, score, confidence]

                    self.size_locations.append(size)
                    self.score_locations.append(score)
                    self.confidence_locations.append(confidence)
            point_properties = {
                "score": np.array(self.score_locations),
                "confidence": np.array(self.confidence_locations),
                "size": np.array(self.size_locations),
            }
            text_properties = {
                "text": event_name
                + ": {score:.5f}"
                + "\n"
                + "Confidence"
                + ": {confidence:.5f}"
                + "\n"
                + "Size"
                + ": {size:.5f}",
                "anchor": "upper_left",
                "translation": [-5, 0],
                "size": 12,
                "color": "pink",
            }
            name_remove = ("Detections", "Location Map")
            for layer in list(self.viewer.layers):

                if any(name in layer.name for name in name_remove):
                    self.viewer.layers.remove(layer)
            if len(self.score_locations) > 0:
                self.viewer.add_points(
                    self.event_locations,
                    properties=point_properties,
                    text=text_properties,
                    name="Detections" + event_name,
                    face_color=[0] * 4,
                    edge_color="red",
                )

            if segimagedir is not None:
                for layer in list(self.viewer.layers):
                    if isinstance(layer, layers.Labels):
                        self.seg_image = layer.data

                        location_image, self.cell_count = location_map(
                            self.event_locations_dict,
                            self.seg_image,
                            heatmapsteps,
                            display_3d=False,
                        )
                        self.viewer.add_labels(
                            location_image.astype("uint16"),
                            name="Location Map" + imagename,
                        )

            self.cluster_points(nms_space)


def MidSlices(Image, start_project_mid, end_project_mid, axis=1):

    SmallImage = Image.take(
        indices=range(
            Image.shape[axis] // 2 - start_project_mid,
            Image.shape[axis] // 2 + end_project_mid,
        ),
        axis=axis,
    )
    MaxProject = np.amax(SmallImage, axis=axis)

    return MaxProject


def DownsampleData(image, DownsampleFactor):

    if DownsampleFactor != 1:

        print("Downsampling Image in XY by", DownsampleFactor)
        scale_percent = int(100 / DownsampleFactor)  # percent of original size
        width = int(image.shape[2] * scale_percent / 100)
        height = int(image.shape[1] * scale_percent / 100)
        dim = (width, height)
        smallimage = np.zeros([image.shape[0], height, width])
        for i in range(0, image.shape[0]):
            # resize image
            smallimage[i, :] = zoom(image[i, :].astype("float32"), dim)

        return smallimage
    else:

        return image


def PatchGenerator(
    image,
    resultsdir,
    csv_gt,
    number_patches,
    patch_shape,
    size_tminus,
    size_tplus,
    DownsampleFactor=1,
):

    image = DownsampleData(image, DownsampleFactor)
    dataset_gt = pd.read_csv(csv_gt, delimiter=",")

    dataset_gt = dataset_gt.sample(frac=1)
    T_gt = dataset_gt[dataset_gt.keys()[0]][0:]
    Y_gt = dataset_gt[dataset_gt.keys()[1]][0:] / DownsampleFactor
    X_gt = dataset_gt[dataset_gt.keys()[2]][0:] / DownsampleFactor

    listtime_gt = T_gt.tolist()

    listy_gt = Y_gt.tolist()
    listx_gt = X_gt.tolist()
    count = 0
    Data = []
    for i in range(len(listtime_gt)):
        if count > 2 * number_patches:
            break
        time = int(float(listtime_gt[i])) - 1
        y = float(listy_gt[i])
        x = float(listx_gt[i])

        if (
            x > 0.25 * image.shape[2]
            and x < 0.75 * image.shape[2]
            and y > 0.25 * image.shape[1]
            and y < 0.75 * image.shape[1]
        ):
            crop_Xminus = x - int(patch_shape[0] / 2)
            crop_Xplus = x + int(patch_shape[0] / 2)
            crop_Yminus = y - int(patch_shape[1] / 2)
            crop_Yplus = y + int(patch_shape[1] / 2)

            randomy = np.random.randint(
                min(0.25 * image.shape[2], 0.25 * image.shape[1]),
                high=max(0.25 * image.shape[2], 0.25 * image.shape[1]),
            )
            randomx = np.random.randint(
                min(0.25 * image.shape[2], 0.25 * image.shape[1]),
                high=max(0.25 * image.shape[2], 0.25 * image.shape[1]),
            )
            random_crop_Xminus = randomx - int(patch_shape[0] / 2)
            random_crop_Xplus = randomx + int(patch_shape[0] / 2)
            random_crop_Yminus = randomy - int(patch_shape[1] / 2)
            random_crop_Yplus = randomy + int(patch_shape[1] / 2)

            region = (
                slice(int(time - size_tminus), int(time + size_tplus + 1)),
                slice(int(crop_Yminus), int(crop_Yplus)),
                slice(int(crop_Xminus), int(crop_Xplus)),
            )

            random_region = (
                slice(int(time - size_tminus), int(time + size_tplus + 1)),
                slice(int(random_crop_Yminus), int(random_crop_Yplus)),
                slice(int(random_crop_Xminus), int(random_crop_Xplus)),
            )

            crop_image = image[region]
            random_crop_image = image[random_region]
            if (
                crop_image.shape[0] == size_tplus + size_tminus + 1
                and crop_image.shape[1] == patch_shape[1]
                and crop_image.shape[2] == patch_shape[0]
            ):
                Data.append([time, y * DownsampleFactor, x * DownsampleFactor])
                imwrite(
                    resultsdir
                    + "Skeletor"
                    + "T"
                    + str(time)
                    + "Y"
                    + str(y * DownsampleFactor)
                    + "X"
                    + str(x * DownsampleFactor)
                    + ".tif",
                    crop_image.astype("float16"),
                    metadata={"axes": "TYX"},
                )
            count = count + 1
            if (
                random_crop_image.shape[0] == size_tplus + size_tminus + 1
                and random_crop_image.shape[1] == patch_shape[1]
                and random_crop_image.shape[2] == patch_shape[0]
            ):
                Data.append(
                    [
                        time,
                        randomy * DownsampleFactor,
                        randomx * DownsampleFactor,
                    ]
                )
                imwrite(
                    resultsdir
                    + "Skeletor"
                    + "T"
                    + str(time)
                    + "Y"
                    + str(randomy * DownsampleFactor)
                    + "X"
                    + str(randomx * DownsampleFactor)
                    + ".tif",
                    random_crop_image.astype("float16"),
                    metadata={"axes": "TYX"},
                )
            count = count + 1

    writer = csv.writer(open(resultsdir + "/" + ("GTLocator") + ".csv", "w"))
    writer.writerows(Data)


def GetMarkers(image):

    MarkerImage = np.zeros(image.shape)
    waterproperties = measure.regionprops(image)
    Coordinates = [prop.centroid for prop in waterproperties]
    Coordinates = sorted(Coordinates, key=lambda k: [k[0], k[1]])
    coordinates_int = np.round(Coordinates).astype(int)
    MarkerImage[tuple(coordinates_int.T)] = 1 + np.arange(len(Coordinates))

    markers = morphology.dilation(MarkerImage, morphology.disk(2))

    return markers



import collections
import csv
import json
import os
from pathlib import Path
import pandas as pd
import numpy as np
from scipy import spatial
from scipy.ndimage import binary_fill_holes, find_objects
from skimage import measure, morphology
from skimage.measure import label
from skimage.morphology import binary_dilation, dilation, square
from skimage.util import invert as invertimage
from tifffile import imwrite
from tqdm import tqdm
from skimage.measure import regionprops
from tifffile import imread

def location_map(
    event_locations_dict: dict,
    seg_image: np.ndarray,
    heatmapsteps: int,
    display_3d: bool = True,
):
    cell_count = {}
    location_image = np.zeros(seg_image.shape)
    j = 0
    for i in range(seg_image.shape[0]):
        current_seg_image = seg_image[i, :]
        waterproperties = measure.regionprops(current_seg_image)
        indices = [prop.centroid for prop in waterproperties]
        cell_count[int(i)] = len(indices)

        if int(i) in event_locations_dict.keys():
            currentindices = event_locations_dict[int(i)]

            if len(indices) > 0:
                tree = spatial.cKDTree(indices)
                if len(currentindices) > 0:
                    for j in range(0, len(currentindices)):
                        index = currentindices[j]
                        closest_marker_index = tree.query(index)
                        if display_3d:
                            current_seg_label = current_seg_image[
                                int(indices[closest_marker_index[1]][0]),
                                int(indices[closest_marker_index[1]][1]),
                                int(indices[closest_marker_index[1]][2]),
                            ]
                        else:
                            current_seg_label = current_seg_image[
                                int(indices[closest_marker_index[1]][0]),
                                int(indices[closest_marker_index[1]][1]),
                            ]
                        if current_seg_label > 0:
                            all_pixels = np.where(
                                current_seg_image == current_seg_label
                            )
                            all_pixels = np.asarray(all_pixels)
                            if display_3d:
                                for k in range(all_pixels.shape[1]):
                                    location_image[
                                        i,
                                        all_pixels[0, k],
                                        all_pixels[1, k],
                                        all_pixels[2, k],
                                    ] = 1
                            else:
                                for k in range(all_pixels.shape[1]):
                                    location_image[
                                        i, all_pixels[0, k], all_pixels[1, k]
                                    ] = 1

    location_image = average_heat_map(location_image, heatmapsteps)

    return location_image, cell_count


def average_heat_map(image, sliding_window):

    j = 0
    for i in range(image.shape[0]):

        j = j + 1
        if i > 0:
            image[i, :] = np.add(image[i, :], image[i - 1, :])
        if j == sliding_window:
            image[i, :] = np.subtract(image[i, :], image[i - 1, :])
            j = 0
    return image


def remove_big_objects(ar, max_size=6400, connectivity=1, in_place=False):
    out = ar.copy()
    ccs = out

    try:
        component_sizes = np.bincount(ccs.ravel())
    except ValueError:
        raise ValueError(
            "Negative value labels are not supported. Try "
            "relabeling the input with `scipy.ndimage.label` or "
            "`skimage.morphology.label`."
        )

    too_big = component_sizes > max_size
    too_big_mask = too_big[ccs]
    out[too_big_mask] = 0

    return out


def IntergerLabelGen(BinaryImage, Name, savedir):
    InputBinaryImage = BinaryImage.astype("uint8")
    IntegerImage = np.zeros(
        [BinaryImage.shape[0], BinaryImage.shape[1], BinaryImage.shape[2]]
    )
    for i in tqdm(range(0, InputBinaryImage.shape[0])):
        BinaryImageOriginal = InputBinaryImage[i, :]
        Orig = normalizeFloatZeroOne(BinaryImageOriginal)
        InvertedBinaryImage = invertimage(BinaryImageOriginal)
        BinaryImage = normalizeFloatZeroOne(InvertedBinaryImage)
        image = binary_dilation(BinaryImage)
        image = invertimage(image)
        labelclean = label(image)
        labelclean = remove_big_objects(labelclean, max_size=15000)
        AugmentedLabel = dilation(labelclean, selem=square(3))
        AugmentedLabel = np.multiply(AugmentedLabel, Orig)
        IntegerImage[i, :] = AugmentedLabel

    imwrite(savedir + Name + ".tif", IntegerImage.astype("uint16"))


def MarkerToCSV(MarkerImage):
    MarkerImage = MarkerImage.astype("uint16")
    MarkerList = []
    print("Obtaining co-ordinates of markers in all regions")
    for i in range(0, MarkerImage.shape[0]):
        waterproperties = measure.regionprops(MarkerImage, MarkerImage)
        indices = [prop.centroid for prop in waterproperties]
        MarkerList.append([i, indices[0], indices[1]])
    return MarkerList


def load_json(fpath):
    with open(fpath) as f:
        return json.load(f)


def save_json(data, fpath, **kwargs):
    with open(fpath, "w") as f:
        f.write(json.dumps(data, **kwargs))


def normalizeFloatZeroOne(
    x, pmin=1, pmax=99.8, axis=None, eps=1e-20, dtype=np.uint8
):
    """Percentile based Normalization

    Normalize patches of image before feeding into the network

    Parameters
    ----------
    x : np array Image patch
    pmin : minimum percentile value for normalization
    pmax : maximum percentile value for normalization
    axis : axis along which the normalization has to be carried out
    eps : avoid dividing by zero
    """
    mi = np.percentile(x, pmin, axis=axis, keepdims=True)
    ma = np.percentile(x, pmax, axis=axis, keepdims=True)
    return normalize_mi_ma(x, mi, ma, eps=eps, dtype=dtype)


def normalize_mi_ma(x, mi, ma, eps=1e-20, dtype=np.uint8):

    x = x.astype(dtype)
    mi = dtype(mi) if np.isscalar(mi) else mi.astype(dtype, copy=False)
    ma = dtype(ma) if np.isscalar(ma) else ma.astype(dtype, copy=False)
    eps = dtype(eps) if np.isscalar(eps) else eps.astype(dtype, copy=False)

    x = (x - mi) / (ma - mi + eps)

    return x


def generate_membrane_locations(membranesegimage : np.ndarray, csvfile: str, savefile: str):
    
    dataset = pd.read_csv(csvfile, delimiter = ',')
    
    
    writer = csv.writer(open(savefile, "a", newline=""))
    writer.writerow(
                        [
                            "T",
                            "Z",
                            "Y",
                            "X",
                            "Score",
                            "Size",
                            "Confidence",
                        ]
    )
    nrows = len(dataset.columns)
    dict_membrane = {}
    if isinstance(membranesegimage, str):
        membranesegimage = imread(membranesegimage)

    for i in tqdm(range(membranesegimage.shape[0])):
        currentimage = membranesegimage[i,:,:,:]
        properties = measure.regionprops(currentimage) 
        membrane_coordinates = [prop.centroid for prop in properties]
        dict_membrane[i] =  membrane_coordinates
        
    for index, row in dataset.iterrows():
        time = int(row[0])
        z = int(row[1])
        y = int(row[2])
        x = int(row[3])
        index = (z,y,x)
        if time < membranesegimage.shape[0]:
            if nrows > 4:
                        score = row[4]
                        size = row[5]
                        confidence = row[6]
            else:
                        score = 1.0
                        size = 10
                        confidence = 1.0
            membrane_coordinates = dict_membrane[time]
            if len(membrane_coordinates) > 0:
                tree = spatial.cKDTree(membrane_coordinates)  
                distance, nearest_location = tree.query(index)          
                            
                z = int(membrane_coordinates[nearest_location][0])         
                y = membrane_coordinates[nearest_location][1]
                x = membrane_coordinates[nearest_location][2]
                writer.writerow([time, z, y, x, score, size, confidence])


def load_training_data(directory, filename, axes=None, verbose=True):
    """Load training data in .npz format.
    The data file is expected to have the keys 'data' and 'label'
    """
    if directory is not None:
        npzdata = np.load(directory + filename)
    else:
        npzdata = np.load(filename)

    X = npzdata["data"]
    Y = npzdata["label"]
    Z = npzdata["label2"]

    if axes is None:
        axes = npzdata["axes"]
    axes = axes_check_and_normalize(axes)
    assert "C" in axes
    n_images = X.shape[0]
    assert X.shape[0] == Y.shape[0]
    assert 0 < n_images <= X.shape[0]

    X, Y = X[:n_images], Y[:n_images]
    channel = axes_dict(axes)["C"]

    X = move_channel_for_backend(X, channel=channel)

    axes = axes.replace("C", "")  # remove channel
    if backend_channels_last():
        axes = axes + "C"
    else:
        axes = axes[:1] + "C" + axes[1:]

    if verbose:
        ax = axes_dict(axes)
        n_train = len(X)
        image_size = tuple(X.shape[ax[a]] for a in "TZYX" if a in axes)
        n_dim = len(image_size)
        n_channel_in = X.shape[ax["C"]]

        print("number of  images:\t", n_train)

        print("image size (%dD):\t\t" % n_dim, image_size)
        print("axes:\t\t\t\t", axes)
        print("channels in / out:\t\t", n_channel_in)

    return (X, Y, Z), axes


def _raise(e):
    raise e


# https://docs.python.org/3/library/itertools.html#itertools-recipes
def consume(iterator):
    collections.deque(iterator, maxlen=0)


def axes_check_and_normalize(
    axes, length=None, disallowed=None, return_allowed=False
):
    """
    S(ample), T(ime), C(hannel), Z, Y, X
    """
    allowed = "STCZYX"
    axes = str(axes).upper()
    consume(
        a in allowed
        or _raise(
            ValueError(
                "invalid axis '{}', must be one of {}.".format(
                    a, list(allowed)
                )
            )
        )
        for a in axes
    )
    disallowed is None or consume(
        a not in disallowed or _raise(ValueError("disallowed axis '%s'." % a))
        for a in axes
    )
    consume(
        axes.count(a) == 1
        or _raise(ValueError("axis '%s' occurs more than once." % a))
        for a in axes
    )
    length is None or len(axes) == length or _raise(
        ValueError("axes (%s) must be of length %d." % (axes, length))
    )
    return (axes, allowed) if return_allowed else axes


def axes_dict(axes):
    """
    from axes string to dict
    """
    axes, allowed = axes_check_and_normalize(axes, return_allowed=True)
    return {a: None if axes.find(a) == -1 else axes.find(a) for a in allowed}
    # return collections.namedt


def backend_channels_last():
    import keras.backend as K

    assert K.image_data_format() in ("channels_first", "channels_last")
    return K.image_data_format() == "channels_last"


def move_channel_for_backend(X, channel):
    if backend_channels_last():
        return np.moveaxis(X, channel, -1)
    else:
        return np.moveaxis(X, channel, 1)


def load_full_training_data(directory, filename, axes=None, verbose=True):
    """Load training data in .npz format.
    The data file is expected to have the keys 'data' and 'label'
    """

    if directory is not None:
        npzdata = np.load(directory + filename)
    else:
        npzdata = np.load(filename)

    X = npzdata["data"]
    Y = npzdata["label"]

    if axes is None:
        axes = npzdata["axes"]
    axes = axes_check_and_normalize(axes)
    assert "C" in axes
    n_images = X.shape[0]
    assert X.shape[0] == Y.shape[0]
    assert 0 < n_images <= X.shape[0]

    X, Y = X[:n_images], Y[:n_images]
    channel = axes_dict(axes)["C"]

    X = move_channel_for_backend(X, channel=channel)

    axes = axes.replace("C", "")  # remove channel
    if backend_channels_last():
        axes = axes + "C"
    else:
        axes = axes[:1] + "C" + axes[1:]

    if verbose:
        ax = axes_dict(axes)
        n_train = len(X)
        image_size = tuple(X.shape[ax[a]] for a in "TZYX" if a in axes)
        n_dim = len(image_size)
        n_channel_in = X.shape[ax["C"]]

        print("number of  images:\t", n_train)

        print("image size (%dD):\t\t" % n_dim, image_size)
        print("axes:\t\t\t\t", axes)
        print("channels in / out:\t\t", n_channel_in)

    return (X, Y), axes


def pad_timelapse(image, pad_width):

    zero_pad = np.zeros(
        [
            image.shape[0],
            image.shape[1] + pad_width[0],
            image.shape[2] + pad_width[1],
        ]
    )
    for i in range(0, image.shape[0]):

        zero_pad[i, :, :] = np.pad(
            image[i, :, :],
            (
                (pad_width[0] // 2, pad_width[0] // 2),
                (pad_width[1] // 2, pad_width[1] // 2),
            ),
            mode="edge",
        )

    return zero_pad


def time_pad(image, TimeFrames):
    time = image.shape[0]

    timeextend = time

    while timeextend % TimeFrames != 0:
        timeextend = timeextend + 1

    extendimage = np.zeros([timeextend, image.shape[1], image.shape[2]])

    extendimage[0:time, :, :] = image

    return extendimage


def chunk_list(image, patchshape, stride, pair):
    rowstart = pair[0]
    colstart = pair[1]

    endrow = rowstart + patchshape[0]
    endcol = colstart + patchshape[1]

    if endrow > image.shape[1]:
        endrow = image.shape[1]
    if endcol > image.shape[2]:
        endcol = image.shape[2]

    region = (
        slice(0, image.shape[0]),
        slice(rowstart, endrow),
        slice(colstart, endcol),
    )

    # The actual pixels in that region.
    patch = image[region]

    return patch, rowstart, colstart


def DensityCounter(MarkerImage, TrainshapeX, TrainshapeY):
    AllDensity = {}

    for i in tqdm(range(0, MarkerImage.shape[0])):
        density = []
        location = []
        currentimage = MarkerImage[i, :].astype("uint16")
        waterproperties = measure.regionprops(currentimage, currentimage)
        indices = [prop.centroid for prop in waterproperties]

        for y, x in indices:

            crop_Xminus = x - int(TrainshapeX / 2)
            crop_Xplus = x + int(TrainshapeX / 2)
            crop_Yminus = y - int(TrainshapeY / 2)
            crop_Yplus = y + int(TrainshapeY / 2)

            region = (
                slice(int(crop_Yminus), int(crop_Yplus)),
                slice(int(crop_Xminus), int(crop_Xplus)),
            )
            crop_image = currentimage[region].astype("uint16")
            if (
                crop_image.shape[0] >= TrainshapeY
                and crop_image.shape[1] >= TrainshapeX
            ):
                waterproperties = measure.regionprops(crop_image, crop_image)

                labels = [prop.label for prop in waterproperties]
                labels = np.asarray(labels)
                density.append(labels.shape[0])
                location.append((int(y), int(x)))
        AllDensity[str(i)] = [density, location]

    return AllDensity


def fill_label_holes(lbl_img, **kwargs):
    """Fill small holes in label image."""

    # TODO: refactor 'fill_label_holes' and 'edt_prob' to share code
    def grow(sl, interior):
        return tuple(
            slice(s.start - int(w[0]), s.stop + int(w[1]))
            for s, w in zip(sl, interior)
        )

    def shrink(interior):
        return tuple(
            slice(int(w[0]), (-1 if w[1] else None)) for w in interior
        )

    objects = find_objects(lbl_img)
    lbl_img_filled = np.zeros_like(lbl_img)
    for i, sl in enumerate(objects, 1):
        if sl is None:
            continue
        interior = [
            (s.start > 0, s.stop < sz) for s, sz in zip(sl, lbl_img.shape)
        ]
        shrink_slice = shrink(interior)
        grown_mask = lbl_img[grow(sl, interior)] == i
        mask_filled = binary_fill_holes(grown_mask, **kwargs)[shrink_slice]
        lbl_img_filled[sl][mask_filled] = i
    return lbl_img_filled


def dilate_label_holes(lbl_img, iterations):
    lbl_img_filled = np.zeros_like(lbl_img)
    for lbl in range(np.min(lbl_img), np.max(lbl_img) + 1):
        mask = lbl_img == lbl
        mask_filled = binary_dilation(mask, iterations=iterations)
        lbl_img_filled[mask_filled] = lbl
    return lbl_img_filled


def MidSlices(Image, start_project_mid, end_project_mid, axis=1):

    SmallImage = Image.take(
        indices=range(
            Image.shape[axis] // 2 - start_project_mid,
            Image.shape[axis] // 2 + end_project_mid,
        ),
        axis=axis,
    )

    MaxProject = np.amax(SmallImage, axis=axis)

    return MaxProject


def MidSlicesSum(Image, start_project_mid, end_project_mid, axis=1):

    SmallImage = Image.take(
        indices=range(
            Image.shape[axis] // 2 - start_project_mid,
            Image.shape[axis] // 2 + end_project_mid,
        ),
        axis=axis,
    )

    MaxProject = np.sum(SmallImage, axis=axis)

    return MaxProject


def GenerateVolumeMarkers(segimage, pad_width):

    ndim = len(segimage.shape)
    # TZYX
    Markers = np.zeros(
        [
            segimage.shape[0],
            segimage.shape[1],
            segimage.shape[2] + pad_width[0],
            segimage.shape[3] + pad_width[1],
        ],
        dtype=np.uint16,
    )

    for i in tqdm(range(0, segimage.shape[0])):

        smallimage = segimage[i, :]
        newsmallimage = pad_timelapse(smallimage, pad_width)
        properties = measure.regionprops(newsmallimage.astype(np.uint16))

        Coordinates = [prop.centroid for prop in properties]
        if len(Coordinates) > 0:
            Coordinates = sorted(Coordinates, key=lambda k: [k[2], k[1], k[0]])
            Coordinates = np.asarray(Coordinates)

            coordinates_int = np.round(Coordinates).astype(int)
            markers_raw = np.zeros_like(newsmallimage)
            markers_raw[tuple(coordinates_int.T)] = 1 + np.arange(
                len(Coordinates)
            )
            if ndim == 4:
                markers = morphology.dilation(
                    markers_raw.astype(np.uint16), morphology.ball(2)
                )

            Markers[i, :] = label(markers.astype(np.uint16))

    return Markers


def GenerateMarkers(
    segimage, start_project_mid=4, end_project_mid=4, pad_width=(0, 0)
):

    ndim = len(segimage.shape)
    Markers = np.zeros(
        [
            segimage.shape[0],
            segimage.shape[-2] + pad_width[0],
            segimage.shape[-1] + pad_width[1],
        ]
    )

    for i in tqdm(range(0, segimage.shape[0])):

        smallimage = segimage[i, :]
        newsmallimage = pad_timelapse(smallimage, pad_width)
        properties = measure.regionprops(newsmallimage.astype("uint16"))

        Coordinates = [prop.centroid for prop in properties]
        if len(Coordinates) > 0:
            if ndim == 3:
                Coordinates = sorted(Coordinates, key=lambda k: [k[1], k[0]])
            if ndim == 4:
                Coordinates = sorted(
                    Coordinates, key=lambda k: [k[2], k[1], k[0]]
                )
            Coordinates = np.asarray(Coordinates)

            coordinates_int = np.round(Coordinates).astype(int)
            markers_raw = np.zeros_like(newsmallimage)
            markers_raw[tuple(coordinates_int.T)] = 1 + np.arange(
                len(Coordinates)
            )
            if ndim == 4:
                markers = morphology.dilation(
                    markers_raw.astype("uint16"), morphology.ball(2)
                )
            if ndim == 3:
                markers = morphology.dilation(
                    markers_raw.astype("uint16"), morphology.disk(2)
                )
            if ndim == 4:
                markers = MidSlices(
                    markers, start_project_mid, end_project_mid, axis=0
                )

            Markers[i, :] = label(markers.astype("uint16"))

    return Markers


def MakeTrees(segimage):
    AllTrees = {}
    print("Creating Dictionary of marker location for fast search")
    for i in tqdm(range(0, segimage.shape[0])):

        indices = []

        currentimage = segimage[i, :].astype("uint16")

        currentimage = currentimage > 0
        currentimage = label(currentimage)
        props = measure.regionprops(currentimage)
        for prop in props:
            indices.append((int(prop.centroid[0]), int(prop.centroid[1])))
        # Comparison between image_max and im to find the coordinates of local maxima

        if len(indices) > 0:
            tree = spatial.cKDTree(indices)

            AllTrees[str(i)] = [tree, indices]

    return AllTrees


def MakeForest(segimage):
    AllForest = {}
    # TZYX is the shape
    print("Creating Dictionary of marker location for fast search")
    for i in tqdm(range(0, segimage.shape[0])):

        indices = []

        currentimage = segimage[i, :].astype("uint16")
        currentimage = currentimage > 0
        currentimage = label(currentimage)
        props = measure.regionprops(currentimage)
        for prop in props:
            indices.append(
                (
                    int(prop.centroid[0]),
                    int(prop.centroid[1]),
                    int(prop.centroid[2]),
                )
            )

        if len(indices) > 0:
            tree = spatial.cKDTree(indices)

            AllForest[str(i)] = [tree, indices]

    return AllForest


def get_max_score_index(scores, threshold=0, top_k=0, descending=True):
    """Get the max scores with corresponding indicies
    Adapted from the OpenCV c++ source in `nms.inl.hpp <https://github.com/opencv/opencv/blob/ee1e1ce377aa61ddea47a6c2114f99951153bb4f/modules/dnn/src/nms.inl.hpp#L33>`__
    :param scores: a list of scores
    :type scores: list
    :param threshold: consider scores higher than this threshold
    :type threshold: float
    :param top_k: return at most top_k scores; if 0, keep all
    :type top_k: int
    :param descending: if True, list is returened in descending order, else ascending
    :returns: a  sorted by score list  of [score, index]
    """
    score_index = []

    # Generate index score pairs
    for i, score in enumerate(scores):
        if (threshold > 0) and (score > threshold):
            score_index.append([score, i])
        else:
            score_index.append([score, i])

    # Sort the score pair according to the scores in descending order
    npscores = np.array(score_index)

    if descending:
        npscores = npscores[npscores[:, 0].argsort()[::-1]]  # descending order
    else:
        npscores = npscores[npscores[:, 0].argsort()]  # ascending order

    if top_k > 0:
        npscores = npscores[0:top_k]

    return npscores.tolist()


def distance(x, y):

    assert len(x) == len(y)
    assert isinstance(x, list) and isinstance(y, list)
    dist = 0
    for i in range(len(x)):
        dist = dist + (x[i] - y[i]) * (x[i] - y[i])
    dist = abs(dist)

    return dist


def compare_function_volume(box1, box2, imagex, imagey, imagez):
    w1, h1, d1 = box1["width"], box1["height"], box1["depth"]
    w2, h2, d2 = box2["width"], box2["height"], box2["depth"]
    x1 = box1["xstart"]
    x2 = box2["xstart"]

    y1 = box1["ystart"]
    y2 = box2["ystart"]

    z1 = box1["zstart"]
    z2 = box2["zstart"]

    xA = max(x1, x2)
    xB = min(x1 + w1, x2 + w2)
    yA = max(y1, y2)
    yB = min(y1 + h1, y2 + h2)
    zA = max(z1, z2)
    zB = min(z1 + d1, z2 + d2)

    if (
        abs(xA - xB) < imagex - 1
        or abs(yA - yB) < imagey - 1
        or abs(zA - zB) < imagez - 1
    ):
        intersect = max(0, xB - xA) * max(0, yB - yA) * max(0, zB - zA)

        area = h2 * w2 * d2 + h1 * w1 * d1 - intersect

        return float(np.true_divide(intersect, area))
    else:
        return -2


def compare_function(box1, box2, imagex, imagey):
    w1, h1 = box1["width"], box1["height"]
    w2, h2 = box2["width"], box2["height"]
    x1 = box1["xstart"]
    x2 = box2["xstart"]

    y1 = box1["ystart"]
    y2 = box2["ystart"]

    xA = max(x1, x2)
    xB = min(x1 + w1, x2 + w2)
    yA = max(y1, y2)
    yB = min(y1 + h1, y2 + h2)

    if abs(xA - xB) < imagex - 1 or abs(yA - yB) < imagey - 1:
        intersect = max(0, xB - xA) * max(0, yB - yA)

        area = h2 * w2 + h1 * w1 - intersect

        return float(np.true_divide(intersect, area))
    else:
        return -2


def compare_function_sec(box1, box2):

    x1center = box1["xcenter"]
    x2center = box2["xcenter"]

    y1center = box1["ycenter"]
    y2center = box2["ycenter"]

    return distance([x1center, y1center], [x2center, y2center])


def compare_function_sec_volume(box1, box2):

    x1center = box1["xcenter"]
    x2center = box2["xcenter"]

    y1center = box1["ycenter"]
    y2center = box2["ycenter"]

    z1center = box1["zcenter"]
    z2center = box2["zcenter"]

    return distance(
        [x1center, y1center, z1center], [x2center, y2center, z2center]
    )


def goodboxes(
    boxes,
    scores,
    nms_threshold,
    score_threshold,
    imagex,
    imagey,
    fidelity=1,
    nms_function="iou",
):

    if len(boxes) == 0:
        return []

    assert len(scores) == len(boxes)
    assert scores is not None
    if scores is not None:
        assert len(scores) == len(boxes)

    boxes = np.array(boxes)

    # if the bounding boxes integers, convert them to floats --
    # this is important since we'll be doing a bunch of divisions
    if boxes.dtype.kind == "i":
        boxes = boxes.astype("float")

    # initialize the list of picked indexes
    pick = []
    Averageboxes = []
    # sort the bounding boxes by the associated scores
    scores = get_max_score_index(scores, score_threshold, 0, False)
    idxs = np.array(scores, np.int32)[:, 1]

    while len(idxs) > 0:
        # grab the last index in the indexes list, add the index
        # value to the list of picked indexes, then initialize
        # the suppression list (i.e. indexes that will be deleted)
        # using the last index
        last = len(idxs) - 1
        i = idxs[last]
        pick.append(i)
        suppress = [last]
        count = 0
        # loop over all indexes in the indexes list
        for pos in range(0, last):
            # grab the current index
            j = idxs[pos]

            if nms_function == "dist":
                overlap = compare_function_sec(boxes[i], boxes[j])
                overlap_veto = nms_threshold * (
                    imagex * imagex + imagey * imagey
                )
                # if there is sufficient overlap, suppress the current bounding box
                if overlap <= overlap_veto:
                    count = count + 1
                    if count >= fidelity:

                        if boxes[i] not in Averageboxes:
                            Averageboxes.append(boxes[i])
                    suppress.append(pos)
            else:
                overlap = compare_function(boxes[i], boxes[j], imagex, imagey)
                overlap_veto = nms_threshold
                if overlap >= overlap_veto:
                    count = count + 1
                    if count >= fidelity:

                        if boxes[i] not in Averageboxes:
                            Averageboxes.append(boxes[i])
                    suppress.append(pos)

        # delete all indexes from the index list that are in the suppression list
        idxs = np.delete(idxs, suppress)
    # return only the indicies of the bounding boxes that were picked
    return Averageboxes


def volumeboxes(
    boxes,
    scores,
    nms_threshold,
    score_threshold,
    imagex,
    imagey,
    imagez,
    nms_function="iou",
):

    if len(boxes) == 0:
        return []

    assert len(scores) == len(boxes)
    assert scores is not None
    if scores is not None:
        assert len(scores) == len(boxes)

    boxes = np.array(boxes)

    # if the bounding boxes integers, convert them to floats --
    # this is important since we'll be doing a bunch of divisions
    if boxes.dtype.kind == "i":
        boxes = boxes.astype("float")

    # initialize the list of picked indexes
    pick = []
    Averageboxes = []
    # sort the bounding boxes by the associated scores
    scores = get_max_score_index(scores, score_threshold, 0, False)
    idxs = np.array(scores, np.int32)[:, 1]

    while len(idxs) > 0:
        # grab the last index in the indexes list, add the index
        # value to the list of picked indexes, then initialize
        # the suppression list (i.e. indexes that will be deleted)
        # using the last index
        last = len(idxs) - 1
        i = idxs[last]
        pick.append(i)
        suppress = [last]
        count = 0
        # loop over all indexes in the indexes list
        for pos in range(0, last):
            # grab the current index
            j = idxs[pos]

            if nms_function == "dist":
                overlap = compare_function_sec_volume(boxes[i], boxes[j])
                overlap_veto = nms_threshold * (
                    imagex * imagex + imagey * imagey + imagez * imagez
                )
                # if there is sufficient overlap, suppress the current bounding box
                if overlap <= overlap_veto:
                    count = count + 1
                    if boxes[i] not in Averageboxes:
                        Averageboxes.append(boxes[i])
                    suppress.append(pos)
            else:
                overlap = compare_function_volume(
                    boxes[i], boxes[j], imagex, imagey, imagez
                )
                overlap_veto = nms_threshold
                if overlap >= overlap_veto:
                    if boxes[i] not in Averageboxes:
                        Averageboxes.append(boxes[i])
                    suppress.append(pos)

        # delete all indexes from the index list that are in the suppression list
        idxs = np.delete(idxs, suppress)
    # return only the indicies of the bounding boxes that were picked
    return Averageboxes


def goldboxes(
    boxes,
    scores,
    nms_threshold,
    score_threshold,
    gridx,
    gridy,
    nms_function="iou",
):

    if len(boxes) == 0:
        return []

    assert len(scores) == len(boxes)
    assert scores is not None
    if scores is not None:
        assert len(scores) == len(boxes)

    boxes = np.array(boxes)

    # if the bounding boxes integers, convert them to floats --
    # this is important since we'll be doing a bunch of divisions
    if boxes.dtype.kind == "i":
        boxes = boxes.astype("float")

    # initialize the list of picked indexes
    Averageboxes = []
    # sort the bounding boxes by the associated scores
    scores = get_max_score_index(scores, score_threshold, 0, False)
    idxs = np.array(scores, np.int32)[:, 1]

    while len(idxs) > 0:
        # grab the last index in the indexes list, add the index
        # value to the list of picked indexes, then initialize
        # the suppression list (i.e. indexes that will be deleted)
        # using the last index
        last = len(idxs) - 1
        i = idxs[last]
        Averageboxes.append(boxes[i])
        suppress = [last]
        # loop over all indexes in the indexes list
        for pos in range(0, last):
            # grab the current index
            j = idxs[pos]

            if nms_function == "dist":
                overlap = compare_function_sec(boxes[i], boxes[j])
                overlap_veto = nms_threshold * (gridx * gridx + gridy * gridy)
                # if there is sufficient overlap, suppress the current bounding box
                if overlap <= overlap_veto:

                    suppress.append(pos)
            else:
                overlap = compare_function(boxes[i], boxes[j], gridx, gridy)
                overlap_veto = nms_threshold
                if overlap >= overlap_veto:
                    suppress.append(pos)
        # delete all indexes from the index list that are in the suppression list
        idxs = np.delete(idxs, suppress)
    # return only the indicies of the bounding boxes that were picked
    return Averageboxes


def simpleaveragenms(
    boxes,
    scores,
    nms_threshold,
    score_threshold,
    event_name,
    gridx,
    gridy,
    nms_function="iou",
):
    if len(boxes) == 0:
        return []

    assert len(scores) == len(boxes)
    assert scores is not None
    if scores is not None:
        assert len(scores) == len(boxes)

    boxes = np.array(boxes)

    # if the bounding boxes integers, convert them to floats --
    # this is important since we'll be doing a bunch of divisions
    if boxes.dtype.kind == "i":
        boxes = boxes.astype("float")

    # initialize the list of picked indexes
    pick = []
    Averageboxes = []
    newbox = None
    # sort the bounding boxes by the associated scores
    scores = get_max_score_index(scores, score_threshold, 0, False)
    idxs = np.array(scores, np.int32)[:, 1]

    while len(idxs) > 0:
        # grab the last index in the indexes list, add the index
        # value to the list of picked indexes, then initialize
        # the suppression list (i.e. indexes that will be deleted)
        # using the last index
        last = len(idxs) - 1
        i = idxs[last]
        pick.append(i)
        suppress = [last]
        # loop over all indexes in the indexes list
        for pos in range(0, last):
            # grab the current index
            j = idxs[pos]

            # compute the ratio of overlap between the two boxes and the area of the second box
            if nms_function == "dist":
                overlap = compare_function_sec(boxes[i], boxes[j])
                overlap_veto = nms_threshold * (gridx * gridx + gridy * gridy)
                if overlap < overlap_veto:
                    boxAscore = boxes[i][event_name]
                    boxAXstart = boxes[i]["xstart"]
                    boxAYstart = boxes[i]["ystart"]
                    boxAXcenter = boxes[i]["xcenter"]
                    boxAYcenter = boxes[i]["ycenter"]
                    boxArealz = boxes[i]["real_z_event"]
                    boxAheight = boxes[i]["height"]
                    boxAwidth = boxes[i]["width"]
                    boxAconfidence = boxes[i]["confidence"]

                    boxBscore = boxes[j][event_name]
                    boxBXstart = boxes[j]["xstart"]
                    boxBYstart = boxes[j]["ystart"]
                    boxBXcenter = boxes[j]["xcenter"]
                    boxBYcenter = boxes[j]["ycenter"]
                    boxBrealz = boxes[j]["real_z_event"]
                    boxBheight = boxes[j]["height"]
                    boxBwidth = boxes[j]["width"]
                    boxBconfidence = boxes[j]["confidence"]

                    meanboxscore = (boxAscore + boxBscore) / 2
                    meanboxXstart = (boxAXstart + boxBXstart) / 2
                    meanboxYstart = (boxAYstart + boxBYstart) / 2
                    meanboxXcenter = (boxAXcenter + boxBXcenter) / 2
                    meanboxYcenter = (boxAYcenter + boxBYcenter) / 2
                    meanboxrealz = (boxArealz + boxBrealz) / 2
                    meanboxheight = (boxAheight + boxBheight) / 2
                    meanboxwidth = (boxAwidth + boxBwidth) / 2
                    meanboxconfidence = (boxAconfidence + boxBconfidence) / 2

                    newbox = {
                        "xstart": meanboxXstart,
                        "ystart": meanboxYstart,
                        "xcenter": meanboxXcenter,
                        "ycenter": meanboxYcenter,
                        "real_z_event": meanboxrealz,
                        "height": meanboxheight,
                        "width": meanboxwidth,
                        "confidence": meanboxconfidence,
                        event_name: meanboxscore,
                    }
                    suppress.append(pos)
            else:
                overlap = compare_function(boxes[i], boxes[j], gridx, gridy)
                overlap_veto = nms_threshold
                if overlap > overlap_veto:
                    boxAscore = boxes[i][event_name]
                    boxAXstart = boxes[i]["xstart"]
                    boxAYstart = boxes[i]["ystart"]
                    boxAXcenter = boxes[i]["xcenter"]
                    boxAYcenter = boxes[i]["ycenter"]
                    boxArealz = boxes[i]["real_z_event"]
                    boxAheight = boxes[i]["height"]
                    boxAwidth = boxes[i]["width"]
                    boxAconfidence = boxes[i]["confidence"]

                    boxBscore = boxes[j][event_name]
                    boxBXstart = boxes[j]["xstart"]
                    boxBYstart = boxes[j]["ystart"]
                    boxBXcenter = boxes[j]["xcenter"]
                    boxBYcenter = boxes[j]["ycenter"]
                    boxBrealz = boxes[j]["real_z_event"]
                    boxBheight = boxes[j]["height"]
                    boxBwidth = boxes[j]["width"]
                    boxBconfidence = boxes[j]["confidence"]

                    meanboxscore = (boxAscore + boxBscore) / 2
                    meanboxXstart = (boxAXstart + boxBXstart) / 2
                    meanboxYstart = (boxAYstart + boxBYstart) / 2
                    meanboxXcenter = (boxAXcenter + boxBXcenter) / 2
                    meanboxYcenter = (boxAYcenter + boxBYcenter) / 2
                    meanboxrealz = (boxArealz + boxBrealz) / 2
                    meanboxheight = (boxAheight + boxBheight) / 2
                    meanboxwidth = (boxAwidth + boxBwidth) / 2
                    meanboxconfidence = (boxAconfidence + boxBconfidence) / 2

                    newbox = {
                        "xstart": meanboxXstart,
                        "ystart": meanboxYstart,
                        "xcenter": meanboxXcenter,
                        "ycenter": meanboxYcenter,
                        "real_z_event": meanboxrealz,
                        "height": meanboxheight,
                        "width": meanboxwidth,
                        "confidence": meanboxconfidence,
                        event_name: meanboxscore,
                    }
                    suppress.append(pos)

            # if there is sufficient overlap, suppress the current bounding box

        if newbox is not None and newbox not in Averageboxes:
            Averageboxes.append(newbox)
            # delete all indexes from the index list that are in the suppression list
        idxs = np.delete(idxs, suppress)
    # return only the indicies of the bounding boxes that were picked

    meanscore = sum(d[event_name] for d in Averageboxes) / max(
        1, len(Averageboxes)
    )
    maxscore = max(d[event_name] for d in Averageboxes)

    box = {
        "real_z_event": meanboxrealz,
        "confidence": meanboxconfidence,
        event_name: meanscore,
        "max_score": maxscore,
    }

    return box, Averageboxes


def save_static(
    key_categories: dict, iou_classedboxes: dict, all_iou_classed_boxes: dict
):

    for (event_name, event_label) in key_categories.items():

        if event_label > 0:

            for (k, v) in iou_classedboxes.items():
                if k == event_name:
                    boxes = v[0]
                    if k in all_iou_classed_boxes.keys():
                        oldboxes = all_iou_classed_boxes[k]
                        boxes = boxes + oldboxes[0]

            all_iou_classed_boxes[event_name] = [boxes]
    return all_iou_classed_boxes


def save_static_csv(
    key_categories: dict, iou_classedboxes: dict, savedir: str
):

    for (event_name, event_label) in key_categories.items():

        if event_label > 0:
            xlocations = []
            ylocations = []
            scores = []
            confidences = []
            tlocations = []
            radiuses = []

            iou_current_event_boxes = iou_classedboxes[event_name][0]
            iou_current_event_boxes = sorted(
                iou_current_event_boxes,
                key=lambda x: x[event_name],
                reverse=True,
            )

            for iou_current_event_box in iou_current_event_boxes:
                xcenter = iou_current_event_box["xcenter"]
                ycenter = iou_current_event_box["ycenter"]
                tcenter = iou_current_event_box["real_time_event"]
                confidence = iou_current_event_box["confidence"]
                score = iou_current_event_box[event_name]
                radius = (
                    np.sqrt(
                        iou_current_event_box["height"]
                        * iou_current_event_box["height"]
                        + iou_current_event_box["width"]
                        * iou_current_event_box["width"]
                    )
                    // 2
                )
                radius = radius
                # Replace the detection with the nearest marker location
                xlocations.append(xcenter)
                ylocations.append(ycenter)
                scores.append(score)
                confidences.append(confidence)
                tlocations.append(tcenter)
                radiuses.append(radius)

            event_count = np.column_stack(
                [
                    tlocations,
                    ylocations,
                    xlocations,
                    scores,
                    radiuses,
                    confidences,
                ]
            )
            event_count = sorted(
                event_count, key=lambda x: x[0], reverse=False
            )
            event_data = []
            csvname = savedir + "/" + event_name + "Location"
            writer = csv.writer(open(csvname + ".csv", "a", newline=""))
            filesize = os.stat(csvname + ".csv").st_size
            if filesize < 1:
                writer.writerow(["T", "Y", "X", "Score", "Size", "Confidence"])
            for line in event_count:
                if line not in event_data:
                    event_data.append(line)
                writer.writerows(event_data)
                event_data = []


def gold_nms(
    heatmap,
    classedboxes,
    event_name,
    iou_threshold,
    event_threshold,
    gridx,
    gridy,
    generate_map,
    nms_function,
):

    sorted_event_box = classedboxes[event_name][0]
    scores = [
        sorted_event_box[i][event_name] for i in range(len(sorted_event_box))
    ]

    good_sorted_event_box = goldboxes(
        sorted_event_box,
        scores,
        iou_threshold,
        event_threshold,
        gridx,
        gridy,
        nms_function=nms_function,
    )
    for iou_current_event_box in sorted_event_box:
        xcenter = iou_current_event_box["xcenter"]
        ycenter = iou_current_event_box["ycenter"]
        tcenter = iou_current_event_box["real_time_event"]
        score = iou_current_event_box[event_name]

        if generate_map:
            for x in range(int(xcenter - 8), int(xcenter + 8)):
                for y in range(int(ycenter - 8), int(ycenter + 8)):
                    if y < heatmap.shape[1] and x < heatmap.shape[2]:
                        heatmap[int(tcenter), int(y), int(x)] = (
                            heatmap[int(tcenter), int(y), int(x)] + score
                        )

    return good_sorted_event_box


def volume_dynamic_nms(
    classedboxes,
    event_name,
    iou_threshold,
    event_threshold,
    gridx,
    gridy,
    gridz,
    nms_function="iou",
):

    sorted_event_box = classedboxes[event_name][0]
    scores = [
        sorted_event_box[i][event_name] for i in range(len(sorted_event_box))
    ]
    best_sorted_event_box = volumeboxes(
        sorted_event_box,
        scores,
        iou_threshold,
        event_threshold,
        gridx,
        gridy,
        gridz,
        nms_function=nms_function,
    )

    return best_sorted_event_box


def dynamic_nms(
    heatmap,
    classedboxes,
    event_name,
    iou_threshold,
    event_threshold,
    gridx,
    gridy,
    fidelity,
    generate_map=True,
    nms_function="iou",
):

    sorted_event_box = classedboxes[event_name][0]
    scores = [
        sorted_event_box[i][event_name] for i in range(len(sorted_event_box))
    ]
    best_sorted_event_box = goodboxes(
        sorted_event_box,
        scores,
        iou_threshold,
        event_threshold,
        gridx,
        gridy,
        fidelity=fidelity,
        nms_function=nms_function,
    )

    if generate_map:
        for iou_current_event_box in sorted_event_box:
            xcenter = iou_current_event_box["xcenter"]
            ycenter = iou_current_event_box["ycenter"]
            tcenter = iou_current_event_box["real_time_event"]
            score = iou_current_event_box[event_name]

            for x in range(int(xcenter - 8), int(xcenter + 8)):
                for y in range(int(ycenter - 8), int(ycenter + 8)):
                    if y < heatmap.shape[1] and x < heatmap.shape[2]:
                        heatmap[int(tcenter), int(y), int(x)] = (
                            heatmap[int(tcenter), int(y), int(x)] + score
                        )

    return best_sorted_event_box


def microscope_dynamic_nms(
    classedboxes,
    event_name,
    iou_threshold,
    event_threshold,
    gridx,
    gridy,
    fidelity,
    nms_function,
):

    sorted_event_box = classedboxes[event_name][0]
    scores = [
        sorted_event_box[i][event_name] for i in range(len(sorted_event_box))
    ]
    best_sorted_event_box = goodboxes(
        sorted_event_box,
        scores,
        iou_threshold,
        event_threshold,
        gridx,
        gridy,
        fidelity,
        nms_function,
    )

    return best_sorted_event_box


def save_dynamic_csv(key_categories, iou_classedboxes, savedir, z=0):

    for (event_name, event_label) in key_categories.items():

        if event_label > 0:
            xlocations = []
            ylocations = []
            zlocations = []
            scores = []
            confidences = []
            tlocations = []
            radiuses = []

            iou_current_event_boxes = iou_classedboxes[event_name][0]
            iou_current_event_boxes = sorted(
                iou_current_event_boxes,
                key=lambda x: x[event_name],
                reverse=True,
            )
            for iou_current_event_box in iou_current_event_boxes:
                xcenter = iou_current_event_box["xcenter"]
                ycenter = iou_current_event_box["ycenter"]
                tcenter = iou_current_event_box["real_time_event"]
                confidence = iou_current_event_box["confidence"]
                score = iou_current_event_box[event_name]
                radius = (
                    np.sqrt(
                        iou_current_event_box["height"]
                        * iou_current_event_box["height"]
                        + iou_current_event_box["width"]
                        * iou_current_event_box["width"]
                    )
                    // 2
                )
                radius = radius

                xlocations.append(xcenter)
                ylocations.append(ycenter)
                scores.append(score)
                confidences.append(confidence)
                tlocations.append(tcenter)
                radiuses.append(radius)
                zlocations.append(z)
            event_count = np.column_stack(
                [
                    tlocations,
                    zlocations,
                    ylocations,
                    xlocations,
                    scores,
                    radiuses,
                    confidences,
                ]
            )
            event_count = sorted(
                event_count, key=lambda x: x[0], reverse=False
            )
            event_data = []
            csvname = savedir + "/" + event_name + "Location"

            writer = csv.writer(open(csvname + ".csv", "a", newline=""))
            filesize = os.stat(csvname + ".csv").st_size

            if filesize < 1:
                writer.writerow(
                    ["T", "Z", "Y", "X", "Score", "Size", "Confidence"]
                )
            for line in event_count:
                if line not in event_data:
                    event_data.append(line)
                writer.writerows(event_data)
                event_data = []


def save_volume(
    key_categories: dict, iou_classedboxes: dict, all_iou_classed_boxes: dict
):

    for (event_name, event_label) in key_categories.items():

        if event_label > 0:

            for (k, v) in iou_classedboxes.items():
                if k == event_name:
                    boxes = v[0]
                    if k in all_iou_classed_boxes.keys():
                        oldboxes = all_iou_classed_boxes[k]
                        boxes = boxes + oldboxes[0]

            all_iou_classed_boxes[event_name] = [boxes]
    return all_iou_classed_boxes


def save_volume_csv(
    key_categories: dict, iou_classedboxes: dict, savedir: str
):

    for (event_name, event_label) in key_categories.items():

        if event_label > 0:
            xlocations = []
            ylocations = []
            zlocations = []
            scores = []
            confidences = []
            tlocations = []
            radiuses = []

            iou_current_event_boxes = iou_classedboxes[event_name][0]
            iou_current_event_boxes = sorted(
                iou_current_event_boxes,
                key=lambda x: x[event_name],
                reverse=True,
            )
            for iou_current_event_box in iou_current_event_boxes:
                xcenter = iou_current_event_box["xcenter"]
                ycenter = iou_current_event_box["ycenter"]
                zcenter = iou_current_event_box["zcenter"]
                tcenter = iou_current_event_box["real_time_event"]
                confidence = iou_current_event_box["confidence"]
                score = iou_current_event_box[event_name]
                radius = (
                    np.sqrt(
                        iou_current_event_box["height"]
                        * iou_current_event_box["height"]
                        + iou_current_event_box["width"]
                        * iou_current_event_box["width"]
                        + iou_current_event_box["depth"]
                        * iou_current_event_box["depth"]
                    )
                    // 3
                )

                xlocations.append(xcenter)
                ylocations.append(ycenter)
                zlocations.append(zcenter)
                scores.append(score)
                confidences.append(confidence)
                tlocations.append(tcenter)
                radiuses.append(radius)

            event_count = np.column_stack(
                [
                    tlocations,
                    zlocations,
                    ylocations,
                    xlocations,
                    scores,
                    radiuses,
                    confidences,
                ]
            )
            event_count = sorted(
                event_count, key=lambda x: x[0], reverse=False
            )
            event_data = []
            csvname = savedir + "/" + f"pred_{event_name}_locations"

            writer = csv.writer(open(csvname + ".csv", "a", newline=""))
            filesize = os.stat(csvname + ".csv").st_size

            if filesize < 1:
                writer.writerow(
                    ["T", "Z", "Y", "X", "Score", "Size", "Confidence"]
                )
            for line in event_count:
                if line not in event_data:
                    event_data.append(line)
                writer.writerows(event_data)
                event_data = []


def yoloprediction(
    sy,
    sx,
    time_prediction,
    stride,
    inputtime,
    config,
    key_categories,
    key_cord,
    nboxes,
    mode,
    event_type,
    marker_tree=None,
):
    LocationBoxes = []
    j = 0
    k = 1
    while True:
        j = j + 1
        if j > time_prediction.shape[1]:
            j = 1
            k = k + 1

        if k > time_prediction.shape[0]:
            break
        Classybox = predictionloop(
            j,
            k,
            sx,
            sy,
            nboxes,
            stride,
            time_prediction,
            config,
            key_categories,
            key_cord,
            inputtime,
            mode,
            event_type,
            marker_tree=marker_tree,
        )
        # Append the box and the maximum likelehood detected class
        if Classybox is not None:
            LocationBoxes.append(Classybox)

    return LocationBoxes


def volumeyoloprediction(
    sz,
    sy,
    sx,
    time_prediction,
    stride,
    inputtime,
    config,
    key_categories,
    key_cord,
    nboxes,
    mode,
    event_type,
    marker_tree=None,
):
    LocationBoxes = []
    j = 0

    k = 1
    for i in range(time_prediction.shape[0]):
        while True:
            j = j + 1

            if j > time_prediction.shape[1]:
                j = 1

                k = k + 1

            if k > time_prediction.shape[2]:
                break
            Classybox = volumepredictionloop(
                i,
                j,
                k,
                sz,
                sy,
                sx,
                nboxes,
                stride,
                time_prediction,
                config,
                key_categories,
                key_cord,
                inputtime,
                mode,
                event_type,
                marker_tree=marker_tree,
            )
            # Append the box and the maximum likelehood detected class
            if Classybox is not None:
                LocationBoxes.append(Classybox)

    return LocationBoxes


def volumepredictionloop(
    i,
    j,
    k,
    sz,
    sy,
    sx,
    nboxes,
    stride,
    time_prediction,
    config,
    key_categories,
    key_cord,
    inputtime,
    mode,
    event_type,
    marker_tree=None,
):
    total_classes = len(key_categories)
    total_coords = len(key_cord)
    y = (k - 1) * stride
    x = (j - 1) * stride
    z = (i - 1) * stride
    prediction_vector = time_prediction[i - 1, k - 1, j - 1, :]

    xstart = x + sx
    ystart = y + sy
    zstart = z + sz
    Class = {}
    # Compute the probability of each class
    for (event_name, event_label) in key_categories.items():
        Class[event_name] = prediction_vector[event_label]

    xcentermean = 0
    ycentermean = 0

    zcentermean = 0
    xcenterrawmean = 0
    ycenterrawmean = 0
    zcenterrawmean = 0

    widthmean = 0
    heightmean = 0
    depthmean = 0

    zcenter = 0
    confidencemean = 0
    trainshapex = config["imagex"]
    trainshapey = config["imagey"]
    trainshapez = config["imagez"]
    zcenterraw = 0
    for b in range(0, nboxes):
        xcenter = (
            xstart
            + prediction_vector[total_classes + config["x"] + b * total_coords]
            * trainshapex
        )
        ycenter = (
            ystart
            + prediction_vector[total_classes + config["y"] + b * total_coords]
            * trainshapey
        )
        zcenter = (
            zstart
            + prediction_vector[total_classes + config["z"] + b * total_coords]
            * trainshapez
        )
        xcenterraw = prediction_vector[
            total_classes + config["x"] + b * total_coords
        ]
        ycenterraw = prediction_vector[
            total_classes + config["y"] + b * total_coords
        ]
        zcenterraw = prediction_vector[
            total_classes + config["z"] + b * total_coords
        ]
        try:
            height = (
                prediction_vector[
                    total_classes + config["h"] + b * total_coords
                ]
                * trainshapex
            )
            width = (
                prediction_vector[
                    total_classes + config["w"] + b * total_coords
                ]
                * trainshapey
            )
            depth = (
                prediction_vector[
                    total_classes + config["d"] + b * total_coords
                ]
                * trainshapez
            )
        except ValueError:
            height = trainshapey
            width = trainshapex
            depth = trainshapez
        if event_type == "dynamic" and mode == "detection":
            confidence = prediction_vector[
                total_classes + config["c"] + b * total_coords
            ]

        if mode == "prediction":
            confidence = 1

        if event_type == "static":

            confidence = prediction_vector[
                total_classes + config["c"] + b * total_coords
            ]

        xcentermean = xcentermean + xcenter
        ycentermean = ycentermean + ycenter
        zcentermean = zcentermean + zcenter

        heightmean = heightmean + height
        widthmean = widthmean + width
        depthmean = depthmean + depth

        confidencemean = confidencemean + confidence

        xcenterrawmean = xcenterrawmean + xcenterraw
        ycenterrawmean = ycenterrawmean + ycenterraw
        zcenterrawmean = zcenterrawmean + zcenterraw

    xcentermean = xcentermean / nboxes
    ycentermean = ycentermean / nboxes
    zcentermean = zcentermean / nboxes

    heightmean = heightmean / nboxes
    widthmean = widthmean / nboxes
    depthmean = depthmean / nboxes

    confidencemean = confidencemean / nboxes
    xcenterrawmean = xcenterrawmean / nboxes
    ycenterrawmean = ycenterrawmean / nboxes
    zcenterrawmean = zcenterrawmean / nboxes

    classybox = {}

    box = None
    if event_type == "dynamic":
        if mode == "detection":
            real_time_event = inputtime
        if mode == "prediction":
            real_time_event = int(inputtime)

        # Compute the box vectors
        if marker_tree is not None:
            nearest_location = get_nearest_volume(
                marker_tree,
                zcentermean,
                ycentermean,
                xcentermean,
                real_time_event,
            )
            if nearest_location is not None:
                zcentermean, ycentermean, xcentermean = nearest_location
        # Correct for zero padding

        box = {
            "xstart": xstart,
            "ystart": ystart,
            "zstart": zstart,
            "xcenterraw": xcenterrawmean,
            "ycenterraw": ycenterrawmean,
            "zcenterraw": zcenterrawmean,
            "xcenter": xcentermean,
            "ycenter": ycentermean,
            "zcenter": zcentermean,
            "real_time_event": real_time_event,
            "height": heightmean,
            "width": widthmean,
            "depth": depthmean,
            "confidence": confidencemean,
        }

    if event_type == "static":
        real_time_event = int(inputtime)
        if marker_tree is not None:
            nearest_location = get_nearest_volume(
                marker_tree,
                zcentermean,
                ycentermean,
                xcentermean,
                real_time_event,
            )
            if nearest_location is not None:
                zcentermean, ycentermean, xcentermean = nearest_location

        box = {
            "xstart": xstart,
            "ystart": ystart,
            "zstart": zstart,
            "xcenterraw": xcenterrawmean,
            "ycenterraw": ycenterrawmean,
            "zcenterraw": zcenterrawmean,
            "xcenter": xcentermean,
            "ycenter": ycentermean,
            "zcenter": zcentermean,
            "real_time_event": real_time_event,
            "height": heightmean,
            "width": widthmean,
            "depth": depthmean,
            "confidence": confidencemean,
        }

    if box is not None:
        # Make a single dict object containing the class and the box vectors return also the max prob label
        for d in [Class, box]:
            classybox.update(d)

        return classybox


def focyoloprediction(
    sy, sx, z_prediction, stride, inputz, config, key_categories
):
    LocationBoxes = []
    j = 0
    k = 1
    while True:
        j = j + 1
        if j > z_prediction.shape[2]:
            j = 1
            k = k + 1

        if k > z_prediction.shape[1]:
            break
        Classybox = focpredictionloop(
            j, k, sx, sy, stride, z_prediction, config, key_categories, inputz
        )
        # Append the box and the maximum likelehood detected class
        if Classybox is not None:
            LocationBoxes.append(Classybox)
    return LocationBoxes


def predictionloop(
    j,
    k,
    sx,
    sy,
    nboxes,
    stride,
    time_prediction,
    config,
    key_categories,
    key_cord,
    inputtime,
    mode,
    event_type,
    marker_tree=None,
):
    total_classes = len(key_categories)
    total_coords = len(key_cord)
    y = (k - 1) * stride
    x = (j - 1) * stride

    prediction_vector = time_prediction[k - 1, j - 1, :]

    xstart = x + sx
    ystart = y + sy
    Class = {}
    # Compute the probability of each class
    for (event_name, event_label) in key_categories.items():
        Class[event_name] = prediction_vector[event_label]

    xcentermean = 0
    ycentermean = 0
    tcentermean = 0
    xcenterrawmean = 0
    ycenterrawmean = 0
    tcenterrawmean = 0
    boxtcentermean = 0
    widthmean = 0
    heightmean = 0
    boxtstartmean = 0
    boxtstart = inputtime
    tcenter = 0
    boxtcenter = 0
    confidencemean = 0
    trainshapex = config["imagex"]
    trainshapey = config["imagey"]
    tcenterraw = 0
    for b in range(0, nboxes):
        xcenter = (
            xstart
            + prediction_vector[total_classes + config["x"] + b * total_coords]
            * trainshapex
        )
        ycenter = (
            ystart
            + prediction_vector[total_classes + config["y"] + b * total_coords]
            * trainshapey
        )
        xcenterraw = prediction_vector[
            total_classes + config["x"] + b * total_coords
        ]
        ycenterraw = prediction_vector[
            total_classes + config["y"] + b * total_coords
        ]

        try:
            height = (
                prediction_vector[
                    total_classes + config["h"] + b * total_coords
                ]
                * trainshapex
            )
            width = (
                prediction_vector[
                    total_classes + config["w"] + b * total_coords
                ]
                * trainshapey
            )
        except ValueError:
            height = trainshapey
            width = trainshapex
        if event_type == "dynamic" and mode == "detection":
            time_frames = config["size_tminus"] + config["size_tplus"] + 1
            tcenter = int(
                inputtime
                + (
                    prediction_vector[
                        total_classes + config["t"] + b * total_coords
                    ]
                    * time_frames
                )
            )
            tcenterraw = prediction_vector[
                total_classes + config["t"] + b * total_coords
            ]
            boxtcenter = int(
                prediction_vector[
                    total_classes + config["t"] + b * total_coords
                ]
            )
            boxtstart = inputtime
            confidence = prediction_vector[
                total_classes + config["c"] + b * total_coords
            ]

        if mode == "prediction":
            confidence = 1

        if event_type == "static":

            tcenter = int(inputtime)
            tcenterraw = 1
            boxtstart = inputtime
            confidence = prediction_vector[
                total_classes + config["c"] + b * total_coords
            ]

        xcentermean = xcentermean + xcenter
        ycentermean = ycentermean + ycenter
        heightmean = heightmean + height
        widthmean = widthmean + width
        confidencemean = confidencemean + confidence
        tcentermean = tcentermean + tcenter
        boxtstartmean = boxtstartmean + boxtstart
        boxtcentermean = boxtcentermean + boxtcenter
        xcenterrawmean = xcenterrawmean + xcenterraw
        ycenterrawmean = ycenterrawmean + ycenterraw
        tcenterrawmean = tcenterrawmean + tcenterraw
    xcentermean = xcentermean / nboxes
    ycentermean = ycentermean / nboxes
    heightmean = heightmean / nboxes
    widthmean = widthmean / nboxes
    confidencemean = confidencemean / nboxes
    tcentermean = tcentermean / nboxes
    boxtcentermean = boxtcentermean / nboxes
    xcenterrawmean = xcenterrawmean / nboxes
    ycenterrawmean = ycenterrawmean / nboxes
    tcenterrawmean = tcenterrawmean / nboxes
    boxtstartmean = boxtstartmean / nboxes

    classybox = {}

    box = None
    if event_type == "dynamic":
        if mode == "detection":
            real_time_event = tcentermean
            box_time_event = boxtcentermean
        if mode == "prediction":
            real_time_event = int(inputtime)
            box_time_event = int(inputtime)

        # Compute the box vectors
        if marker_tree is not None:
            nearest_location = get_nearest(
                marker_tree, ycentermean, xcentermean, real_time_event
            )
            if nearest_location is not None:
                ycentermean, xcentermean = nearest_location
        # Correct for zero padding

        box = {
            "xstart": xstart,
            "ystart": ystart,
            "tstart": boxtstartmean,
            "xcenterraw": xcenterrawmean,
            "ycenterraw": ycenterrawmean,
            "tcenterraw": tcenterrawmean,
            "xcenter": xcentermean,
            "ycenter": ycentermean,
            "real_time_event": real_time_event,
            "box_time_event": box_time_event,
            "height": heightmean,
            "width": widthmean,
            "confidence": confidencemean,
        }

    if event_type == "static":
        real_time_event = int(inputtime)
        box_time_event = int(inputtime)
        if marker_tree is not None:
            nearest_location = get_nearest(
                marker_tree, ycentermean, xcentermean, real_time_event
            )
            if nearest_location is not None:
                ycentermean, xcentermean = nearest_location

        box = {
            "xstart": xstart,
            "ystart": ystart,
            "tstart": boxtstartmean,
            "xcenterraw": xcenterrawmean,
            "ycenterraw": ycenterrawmean,
            "tcenterraw": tcenterrawmean,
            "xcenter": xcentermean,
            "ycenter": ycentermean,
            "real_time_event": real_time_event,
            "box_time_event": box_time_event,
            "height": heightmean,
            "width": widthmean,
            "confidence": confidencemean,
        }

    if box is not None:
        # Make a single dict object containing the class and the box vectors return also the max prob label
        for d in [Class, box]:
            classybox.update(d)

        return classybox


def focpredictionloop(
    j, k, sx, sy, stride, time_prediction, config, key_categories, inputz
):
    total_classes = len(key_categories)
    y = (k - 1) * stride
    x = (j - 1) * stride
    prediction_vector = time_prediction[0, k - 1, j - 1, :]

    xstart = x + sx
    ystart = y + sy
    Class = {}
    # Compute the probability of each class
    for (event_name, event_label) in key_categories.items():
        Class[event_name] = prediction_vector[event_label]

    trainshapex = config["imagex"]
    trainshapey = config["imagey"]

    xcentermean = xstart + 0.5 * trainshapex
    ycentermean = ystart + 0.5 * trainshapey

    heightmean = trainshapey
    widthmean = trainshapex
    zcentermean = int(inputz)
    confidencemean = prediction_vector[total_classes + config["c"]]

    max_prob_label = np.argmax(prediction_vector[:total_classes])

    if max_prob_label >= 0:

        real_z_event = zcentermean

        # Compute the box vectors
        box = {
            "xstart": xstart,
            "ystart": ystart,
            "xcenter": xcentermean,
            "ycenter": ycentermean,
            "real_z_event": real_z_event,
            "height": heightmean,
            "width": widthmean,
            "confidence": confidencemean,
        }

        # Make a single dict object containing the class and the box vectors return also the max prob label
        classybox = {}
        for d in [Class, box]:
            classybox.update(d)

        return classybox


def get_nearest(marker_tree, ycenter, xcenter, tcenter):
    location = (ycenter, xcenter)
    tree, indices = marker_tree[str(int(round(tcenter)))]
    distance, nearest_location = tree.query(location)

    if distance <= 30:
        nearest_location = int(indices[nearest_location][0]), int(
            indices[nearest_location][1]
        )
        return nearest_location[0], nearest_location[1]
    else:
        return None


def get_nearest_volume(marker_tree, zcenter, ycenter, xcenter, tcenter):
    location = (zcenter, ycenter, xcenter)
    tree, indices = marker_tree[str(int(round(tcenter)))]
    distance, nearest_location = tree.query(location)

    if distance <= 30:
        nearest_location = (
            int(indices[nearest_location][0]),
            int(indices[nearest_location][1]),
            int(indices[nearest_location][2]),
        )
        return nearest_location[0], nearest_location[1], nearest_location[2]
    else:
        return None


def save_labelimages(save_dir, image, axes, fname, Name):
    imwrite((save_dir + Name + ".tif"), image)


def Distance(centerAx, centerAy, centerBx, centerBy):

    distance = (centerAx - centerBx) * (centerAx - centerBx) + (
        centerAy - centerBy
    ) * (centerAy - centerBy)

    return distance


def Genericdist(vec1, vec2):

    distance = 0
    for i in range(0, len(vec1)):
        distance = distance + (vec1[i] - vec2[i]) * (vec1[i] - vec2[i])
    return distance


def save_csv(save_dir, Event_Count, Name):
    Event_data = []

    Path(save_dir).mkdir(exist_ok=True)

    for line in Event_Count:
        Event_data.append(line)
    writer = csv.writer(
        open(save_dir + "/" + (Name) + ".csv", "w", newline="")
    )
    writer.writerows(Event_data)



from oneat.NEATUtils import plotters
import numpy as np
from oneat.NEATUtils import utils
from oneat.NEATUtils.utils import save_volume, pad_timelapse, get_nearest_volume,  load_json, volumeyoloprediction, normalizeFloatZeroOne, GenerateVolumeMarkers, MakeForest,save_volume_csv, volume_dynamic_nms
from keras import callbacks
import os
import sys
import datetime
import tensorflow as tf
from tqdm import tqdm
from oneat.NEATModels import nets
from oneat.NEATModels.nets import Concat
from oneat.NEATModels.loss import volume_yolo_loss
from oneat.pretrained import get_registered_models, get_model_details, get_model_instance
from pathlib import Path
from tensorflow.keras.models import load_model
from tifffile import imread


class NEATDenseVollNet(object):
    """
    Parameters
    ----------
    
    NpzDirectory : Specify the location of npz file containing the training data with movies and labels
    
    TrainModelName : Specify the name of the npz file containing training data and labels
    
    ValidationModelName :  Specify the name of the npz file containing validation data and labels
    
    categories : Number of action classes
    
    Categories_Name : List of class names and labels
    
    model_dir : Directory location where trained model weights are to be read or written from
    
    model_keras : The model as it appears as a Keras function
    
    model : If re-training model = model_dir else None as default
    
    
    epochs :  Number of training epochs, 55 by default
    
    batch_size : batch_size to be used for training, 20 by default
    
    
    
    """

    def __init__(self, config, model_dir, catconfig, cordconfig):

        self.config = config
        self.catconfig = catconfig
        self.cordconfig = cordconfig
        self.model_dir = model_dir
        self.key_cord = self.cordconfig
        self.categories = len(self.catconfig)
        self.key_categories = self.catconfig
        if self.config != None:
            self.npz_directory = config.npz_directory
            self.npz_name = config.npz_name
            self.npz_val_name = config.npz_val_name
            self.key_categories = config.key_categories
            self.stage_number = config.stage_number
            self.last_conv_factor = 2 ** (self.stage_number - 1)
            self.show = config.show
            self.key_cord = config.key_cord
            self.box_vector = len(config.key_cord)
            self.categories = len(config.key_categories)
            self.depth = config.depth
            self.start_kernel = config.start_kernel
            self.mid_kernel = config.mid_kernel
            self.learning_rate = config.learning_rate
            self.epochs = config.epochs
            self.startfilter = config.startfilter
            self.batch_size = config.batch_size
            self.multievent = config.multievent
            self.imagex = config.imagex
            self.imagey = config.imagey
            self.imagez = config.imagez
            self.imaget = config.size_tminus + config.size_tplus + 1
            self.size_tminus = config.size_tminus
            self.size_tplus = config.size_tplus

            self.nboxes = config.nboxes
            self.gridx = 1
            self.gridy = 1
            self.gridz = 1
            self.yolo_v0 = config.yolo_v0
            self.yolo_v1 = config.yolo_v1
            self.yolo_v2 = config.yolo_v2
            self.stride = config.stride
        if self.config == None:

            self.config = load_json(os.path.join(self.model_dir , 'parameters.json'))
            

            self.npz_directory = self.config['npz_directory']
            self.npz_name = self.config['npz_name']
            self.npz_val_name = self.config['npz_val_name']
            
            self.box_vector = self.config['box_vector']
            self.show = self.config['show']
            
            self.depth = self.config['depth']
            self.start_kernel = self.config['start_kernel']
            self.mid_kernel = self.config['mid_kernel']
            self.learning_rate = self.config['learning_rate']
            self.epochs = self.config['epochs']
            self.startfilter = self.config['startfilter']
            self.batch_size = self.config['batch_size']
            self.multievent = self.config['multievent']
            self.imagex = self.config['imagex']
            self.imagey = self.config['imagey']
            self.imagez = self.config['imagez']
            self.imaget = self.config['size_tminus'] + self.config['size_tplus'] + 1
            self.size_tminus = self.config['size_tminus']
            self.size_tplus = self.config['size_tplus']
            self.nboxes = self.config['nboxes']
            self.stage_number = self.config['stage_number']
            self.last_conv_factor = 2 ** (self.stage_number - 1)
            self.gridx = 1
            self.gridy = 1
            self.gridz = 1
            self.yolo_v0 = self.config['yolo_v0']
            self.yolo_v1 = self.config['yolo_v1']
            self.yolo_v2 = self.config['yolo_v2']
            self.stride = self.config['stride']

        self.X = None
        self.Y = None
        self.axes = None
        self.X_val = None
        self.Y_val = None
        self.Trainingmodel = None
        self.Xoriginal = None
        self.Xoriginal_val = None

        self.model_keras = nets.DenseVollNet

        self.last_activation = 'softmax'
        self.entropy = 'notbinary'
        self.yolo_loss = volume_yolo_loss(self.categories, self.gridx, self.gridy, self.gridz, self.nboxes,
                                          self.box_vector, self.entropy)

    @classmethod   
    def local_from_pretrained(cls, name_or_alias=None):
           try:
               print(cls)
               get_model_details(cls, name_or_alias, verbose=True)
               return get_model_instance(cls, name_or_alias)
           except ValueError:
               if name_or_alias is not None:
                   print("Could not find model with name or alias '%s'" % (name_or_alias), file=sys.stderr)
                   sys.stderr.flush()
               get_registered_models(cls, verbose=True)  

    def loadData(self):

        #NTZYX shape is the input
        (X, Y), axes = utils.load_full_training_data(self.npz_directory, self.npz_name, verbose=True)

        (X_val, Y_val), axes = utils.load_full_training_data(self.npz_directory, self.npz_val_name, verbose=True)

        self.Xoriginal = X
        self.Xoriginal_val = X_val
        self.X = X
        self.X =  tf.reshape(self.X, (self.X.shape[0], self.X.shape[2], self.X.shape[3], self.X.shape[4], self.X.shape[1]))
        self.Y = Y[:, :, 0]
        self.X_val = X_val
        self.X_val = tf.reshape(self.X_val, (self.X_val.shape[0], self.X_val.shape[2], self.X_val.shape[3], self.X_val.shape[4], self.X_val.shape[1]))
        self.Y_val = Y_val[:, :, 0]

        self.axes = axes
        self.Y = self.Y.reshape((self.Y.shape[0],1, 1, 1, self.Y.shape[1]))
        self.Y_val = self.Y_val.reshape((self.Y_val.shape[0],1, 1, 1, self.Y_val.shape[1]))

    def TrainModel(self):

        #ZYXT
        input_shape = (self.X.shape[1], self.X.shape[2], self.X.shape[3], self.X.shape[4])
        print(self.X.shape)
        print(input_shape)
        Path(self.model_dir).mkdir(exist_ok=True)

  
        Y_rest = self.Y[:, :, :, :, self.categories:]
        print(Y_rest.shape)

        

        dummyY = np.zeros(
            [self.Y.shape[0], self.Y.shape[1], self.Y.shape[2], self.Y.shape[3], self.categories + self.nboxes * self.box_vector])
        dummyY[:,:, :, :, :self.Y.shape[-1]] = self.Y

        dummyY_val = np.zeros([self.Y_val.shape[0], self.Y_val.shape[1], self.Y_val.shape[2], self.Y_val.shape[3],
                               self.categories + self.nboxes * self.box_vector])
        dummyY_val[:,:, :, :, :self.Y_val.shape[-1]] = self.Y_val

        for b in range(1, self.nboxes):
            dummyY[:,:, :, :, self.categories + b * self.box_vector:self.categories + (b + 1) * self.box_vector] = self.Y[
                                                                                                                 :, :,:,
                                                                                                                 :,
                                                                                                                 self.categories: self.categories + self.box_vector]
            dummyY_val[:, :, :,:,
            self.categories + b * self.box_vector:self.categories + (b + 1) * self.box_vector] = self.Y_val[:, :, :,:,
                                                                                                 self.categories: self.categories + self.box_vector]

        self.Y = dummyY
        self.Y_val = dummyY_val
        print(self.Y.shape)

        self.Trainingmodel = self.model_keras(input_shape, self.categories, 
                                              box_vector=Y_rest.shape[-1], yolo_loss = self.yolo_loss, nboxes=self.nboxes,
                                              stage_number=self.stage_number,
                                              depth=self.depth, start_kernel=self.start_kernel,
                                              mid_kernel=self.mid_kernel, 
                                              startfilter=self.startfilter, input_model=self.model_dir,
                                              last_activation=self.last_activation)

        sgd = tf.keras.optimizers.Adam(learning_rate=self.learning_rate)
        self.Trainingmodel.compile(optimizer=sgd, loss=self.yolo_loss, metrics=['accuracy'])
        self.Trainingmodel.summary()
       
   
        # Keras callbacks
        lrate = callbacks.ReduceLROnPlateau(monitor='loss', factor=0.1, patience=4, verbose=1)
        hrate = callbacks.History()
        srate = callbacks.ModelCheckpoint(self.model_dir, monitor='loss', verbose=1,
                                          save_best_only=False, save_weights_only=False, mode='auto', save_freq=1)
        prate = plotters.PlotVolumeHistory(self.Trainingmodel, self.X_val, self.Y_val, self.key_categories, self.key_cord,
                                     self.gridx, self.gridy, self.gridz, plot=self.show, nboxes=self.nboxes)
        
        log_dir = "logs/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        tensorboard_callback = callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1)
        # Train the model and save as a h5 file
        self.Trainingmodel.fit(self.X, self.Y, batch_size=self.batch_size,
                               epochs=self.epochs, validation_data=(self.X_val, self.Y_val), shuffle=True,
                               callbacks=[lrate, hrate, srate, prate, tensorboard_callback])


        self.Trainingmodel.save(self.model_dir)
    """
    The input image and seg image are numpy arrays that have to be read prior to being loaded in the function
    """
    def get_markers(self, 
                    segimage : np.ndarray):

        self.segimage = segimage.astype(np.uint16)
        print('Obtaining Markers')
        self.pad_width = (self.config['imagey'], self.config['imagex'])
        self.markers = GenerateVolumeMarkers(self.segimage, pad_width = self.pad_width)
        self.marker_tree = MakeForest(self.markers)
        self.segimage = None         

        return self.marker_tree
    
    def predict(self, 
                image : np.ndarray,  
                savedir : str = None, 
                n_tiles : tuple = (1, 1, 1), 
                overlap_percent : float =0.8,
                event_threshold : float = 0.5, 
                event_confidence : float = 0.5, 
                iou_threshold : float = 0.1,  
                dtype : np.dtype = np.uint8,
                marker_tree : dict = None, 
                remove_markers : bool = False, 
                nms_function : str = 'iou', 
                activations : bool = False):
        
        self.dtype = dtype
        self.nms_function = nms_function 
        self.originalimage = image.astype(self.dtype)
        self.ndim = len(self.originalimage.shape)
        self.activations = activations
        self.savedir = savedir
        if self.savedir is not None:
           Path(self.savedir).mkdir(exist_ok=True)
        if len(n_tiles) == 3:   
           n_tiles = (n_tiles[-3], n_tiles[-2], n_tiles[-1])
        else:
            n_tiles = (1,1,1)   
        self.n_tiles = n_tiles
        self.overlap_percent = overlap_percent
        self.iou_threshold = iou_threshold
        self.event_threshold = event_threshold
        self.event_confidence = event_confidence
        self.iou_classedboxes = {}
        self.list_all_boxes = []
        self.all_iou_classedboxes = {}
        self.model = self._build()

        self.marker_tree = marker_tree
        self.remove_markers = remove_markers
        
        #Normalize in volume
        self.originalimage = normalizeFloatZeroOne(self.originalimage, 1, 99.8, dtype = self.dtype)
        if self.remove_markers == True:
            self.generate_maps = False 
            self.image = np.zeros([self.originalimage.shape[0], self.originalimage.shape[1],  self.originalimage.shape[2] + self.pad_width[0], self.originalimage.shape[3] + self.pad_width[1] ])
            for i in range(self.originalimage.shape[0]):
               self.image[i,:] = pad_timelapse(self.originalimage[i,:], self.pad_width)
            
            print(f'zero padded image shape ${self.image.shape}')
            self.first_pass_predict()
            self.second_pass_predict()
        if self.remove_markers == False:
           self.generate_maps = False 
           self.image = np.zeros([self.originalimage.shape[0], self.originalimage.shape[1],  self.originalimage.shape[2] + self.pad_width[0], self.originalimage.shape[3] + self.pad_width[1] ])
           for i in range(self.originalimage.shape[0]):
               self.image[i,:] = pad_timelapse(self.originalimage[i,:], self.pad_width)
            
           print(f'zero padded image shape ${self.image.shape}')
           self.second_pass_predict()
        if self.remove_markers == None:
           self.generate_maps = True 
           self.image = self.originalimage
           self.default_pass_predict() 

   
    def _build(self):
        
        Model =  load_model(self.model_dir,
                                custom_objects={'loss': self.yolo_loss, 'Concat': Concat})  
        
        return Model          
     

    def default_pass_predict(self):
        eventboxes = []
        classedboxes = {}    
        count = 0

        print('Detecting event locations')

        for inputtime in tqdm(range(0, self.image.shape[0])):
                    if inputtime < self.image.shape[0] - self.imaget and inputtime > int(self.imaget)//2:
                                count = count + 1
                                      
                                smallimage = CreateVolume(self.image, self.size_tminus, self.size_tplus, inputtime)
                                
                                # Cut off the region for training movie creation
                                #Break image into tiles if neccessary
                                predictions, allx, ally, allz = self.predict_main(smallimage)
                                #Iterate over tiles
                                for p in range(0,len(predictions)):   
                        
                                  sum_time_prediction = predictions[p]
                                  if sum_time_prediction is not None:
                                     #For each tile the prediction vector has shape N H W Categories + Training Vector labels
                                     for i in range(0, sum_time_prediction.shape[0]):
                                          time_prediction =  sum_time_prediction[i]
                                          boxprediction = volumeyoloprediction(
                                          allz[p],  
                                          ally[p], 
                                          allx[p], 
                                          time_prediction, 
                                          self.stride, inputtime , 
                                          self.config, 
                                          self.key_categories, 
                                          self.key_cord, 
                                          self.nboxes, 'detection', 'dynamic',marker_tree=self.marker_tree)
                                          
                                          if boxprediction is not None:
                                                  eventboxes = eventboxes + boxprediction
                                            
                                for (event_name,event_label) in self.key_categories.items(): 
                                                     
                                                if event_label > 0:
                                                     current_event_box = []
                                                     for box in eventboxes:
                                                       
                                                        event_prob = box[event_name]
                                                        event_confidence = box['confidence']
                                                        if event_prob >= self.event_threshold and event_confidence >= self.event_confidence:
                                                            
                                                            current_event_box.append(box)
                                                     classedboxes[event_name] = [current_event_box]
                                                 
                                self.classedboxes = classedboxes   
                                self.eventboxes =  eventboxes
                                
                                
                                if inputtime > 0 and inputtime%(self.imaget) == 0:
                                    
                                    self.nms()
                                    if self.savedir is not None:
                                       self.to_csv()
                                    if self.activations:
                                       self.to_activations()   
                                    eventboxes = []
                                    classedboxes = {}    
                                    count = 0


    def first_pass_predict(self):
        
        print('Detecting background event locations')
        eventboxes = []
        classedboxes = {}
        remove_candidates = {}
        
        
        for inputtime in tqdm(range(0, self.image.shape[0])):
            if inputtime < self.image.shape[0] - self.imaget and inputtime > int(self.imaget)//2:
                
                remove_candidates_list = []
                smallimage = CreateVolume(self.image, self.size_tminus, self.size_tplus, inputtime)
                # Cut off the region for training movie creation
                # Break image into tiles if neccessary
                predictions, allx, ally, allz = self.predict_main(smallimage)
                # Iterate over tiles
                for p in range(0, len(predictions)):

                    sum_time_prediction = predictions[p]

                    if sum_time_prediction is not None:
                        # For each tile the prediction vector has shape N H W Categories + Training Vector labels
                        for i in range(0, sum_time_prediction.shape[0]):
                            time_prediction = sum_time_prediction[i]
                            boxprediction = volumeyoloprediction(
                                          allz[p],  
                                          ally[p], 
                                          allx[p], 
                                          time_prediction, 
                                          self.stride, inputtime , 
                                          self.config, 
                                          self.key_categories, 
                                          self.key_cord, 
                                          self.nboxes, 'detection', 'dynamic',marker_tree=self.marker_tree)
                                          
                            if boxprediction is not None:
                                eventboxes = eventboxes + boxprediction
            for (event_name, event_label) in self.key_categories.items():
                    
                    if event_label == 0:  
                                current_event_box = []              
                                for box in eventboxes:
                
                                    event_prob = box[event_name]
                                    if event_prob >= 0.5 :

                                        current_event_box.append(box)
                                        classedboxes[event_name] = [current_event_box]

            self.classedboxes = classedboxes
            if len(self.classedboxes) > 0:
                self.fast_nms()
                for (event_name, event_label) in self.key_categories.items():
                    
                    if event_label == 0:
                            iou_current_event_boxes = self.iou_classedboxes[event_name][0]
                            iou_current_event_boxes = sorted(iou_current_event_boxes, key=lambda x: x[event_name], reverse=True)
                            for box in iou_current_event_boxes:
                                    
                                     closest_location = get_nearest_volume(self.marker_tree, box['zcenter'], box['ycenter'], box['xcenter'], box['real_time_event'])
                                     if closest_location is not None:
                                        zcentermean, ycentermean, xcentermean = closest_location
                                        try:
                                            remove_candidates_list = remove_candidates[str(int(box['real_time_event']))]
                                            if (zcentermean, ycentermean, xcentermean ) not in remove_candidates_list:
                                                    remove_candidates_list.append((zcentermean, ycentermean, xcentermean))
                                                    remove_candidates[str(int(box['real_time_event']))] = remove_candidates_list
                                        except:
                                            remove_candidates_list.append((zcentermean, ycentermean, xcentermean))
                                            remove_candidates[str(int(box['real_time_event']))]  = remove_candidates_list

                eventboxes = []
                classedboxes = {}                    
       
        



    def second_pass_predict(self):

        print('Detecting event locations')
        eventboxes = []
        classedboxes = {}
        self.n_tiles = (1,1,1)
        
        for inputtime in tqdm(range(int(self.imaget)//2, self.image.shape[0])):
             if inputtime < self.image.shape[0] - self.imaget:   
                smallimage = CreateVolume(self.image, self.size_tminus, self.size_tplus, inputtime)
                if  str(int(inputtime)) in self.marker_tree:                     
                        tree, location = self.marker_tree[str(int(inputtime))]
                        for i in range(len(location)):
                            crop_xminus = location[i][2]  - int(self.imagex/2) 
                            crop_xplus = location[i][2]  + int(self.imagex/2)  
                            
                            crop_yminus = location[i][1]  - int(self.imagey/2)  
                            crop_yplus = location[i][1]   + int(self.imagey/2)  

                            crop_zminus = location[i][0]  - int(self.imagez/2)  
                            crop_zplus = location[i][0]   + int(self.imagez/2) 
                            region =(slice(0,smallimage.shape[0]),slice(int(crop_zminus), int(crop_zplus)),slice(int(crop_yminus), int(crop_yplus)),
                                slice(int(crop_xminus), int(crop_xplus)))
                            
                            crop_image = smallimage[region] 
                            if crop_image.shape[0] >= self.imaget and  crop_image.shape[1] >= self.imagez and crop_image.shape[2] >= self.imagey and crop_image.shape[3] >= self.imagex:                                                
                                        #Now apply the prediction for counting real events
                                        zcenter = location[i][0]
                                        ycenter = location[i][1]
                                        xcenter = location[i][2]
                                        predictions, allx, ally, allz = self.predict_main(crop_image)
                                        sum_time_prediction = predictions[0]
                                        if sum_time_prediction is not None:
                                            #For each tile the prediction vector has shape N H W Categories + Training Vector labels
                                            time_prediction =  sum_time_prediction[0]
                                            boxprediction = volumeyoloprediction(
                                                            0,  
                                                            0, 
                                                            0, 
                                                            time_prediction, 
                                                            self.stride, inputtime , 
                                                            self.config, 
                                                            self.key_categories, 
                                                            self.key_cord, 
                                                            self.nboxes, 'detection', 'dynamic',marker_tree=self.marker_tree)
                                            if boxprediction is not None and len(boxprediction) > 0 and xcenter - self.pad_width[1]//2 > 0 and ycenter - self.pad_width[0]//2 > 0 and xcenter  < self.image.shape[3] - self.pad_width[1]//2 and ycenter  < self.image.shape[2] - self.pad_width[0]//2 :
                                                    
                                                        
                                                        boxprediction[0]['real_time_event'] = inputtime
                                                        boxprediction[0]['xcenter'] = xcenter  - self.pad_width[1]//2
                                                        boxprediction[0]['ycenter'] = ycenter - self.pad_width[0]//2
                                                        boxprediction[0]['zcenter'] = zcenter 


                                                        boxprediction[0]['xstart'] = boxprediction[0]['xcenter']   - int(self.imagex/2) 
                                                        boxprediction[0]['ystart'] = boxprediction[0]['ycenter']   - int(self.imagey/2)   
                                                        boxprediction[0]['zstart'] = zcenter   - int(self.imagez/2)

                                                        eventboxes = eventboxes + boxprediction
                                                
                                           
                for (event_name,event_label) in self.key_categories.items(): 
                                           
                                        if event_label > 0:
                                             current_event_box = []
                                             for box in eventboxes:
                                                event_prob = box[event_name]
                                                event_confidence = box['confidence']
                                                if event_prob >= self.event_threshold and event_confidence >= self.event_confidence :
                                                    current_event_box.append(box)
                                                    
                                             classedboxes[event_name] = [current_event_box]


                self.classedboxes = classedboxes    
                self.eventboxes =  eventboxes
                self.iou_classedboxes = classedboxes
                if self.savedir is not None:
                   self.to_csv()
                if self.activations:
                    self.to_activations()   
                eventboxes = []
                classedboxes = {}   


    def fast_nms(self):


        best_iou_classedboxes = {}
        self.iou_classedboxes = {}
        
        for (event_name,event_label) in self.key_categories.items():
            if event_label == 0:
               #best_sorted_event_box = self.classedboxes[event_name][0]
               best_sorted_event_box = volume_dynamic_nms(self.classedboxes, event_name, self.iou_threshold, self.event_threshold, self.imagex, self.imagey, self.imagez, nms_function = self.nms_function )

               best_iou_classedboxes[event_name] = [best_sorted_event_box]
                   

        self.iou_classedboxes = best_iou_classedboxes
                              


    def nms(self):

        best_iou_classedboxes = {}
        self.iou_classedboxes = {}
        for (event_name,event_label) in self.key_categories.items():
            if event_label > 0:
               #best_sorted_event_box = self.classedboxes[event_name][0]
               best_sorted_event_box = volume_dynamic_nms(self.classedboxes, event_name, self.iou_threshold, self.event_threshold, self.imagex, self.imagey, self.imagez, nms_function = self.nms_function )

               best_iou_classedboxes[event_name] = [best_sorted_event_box]
               
        self.iou_classedboxes = best_iou_classedboxes


    def to_csv(self):
             save_volume_csv( self.key_categories, self.iou_classedboxes, self.savedir)          
 
    def to_activations(self):
             self.all_iou_classedboxes =  save_volume( self.key_categories, self.iou_classedboxes, self.all_iou_classedboxes)

    def overlaptiles(self, sliceregion):

        if self.n_tiles == (1, 1, 1):
            patch = []
            zout = []
            rowout = []
            column = []
          
            patchx = sliceregion.shape[3] // self.n_tiles[2]
            patchy = sliceregion.shape[2] // self.n_tiles[1]
            patchz = sliceregion.shape[1] // self.n_tiles[0]

            patchshape = (patchz, patchy, patchx)
            smallpatch, smallzout,  smallrowout, smallcolumn = chunk_list(sliceregion, patchshape, [0, 0, 0])
            patch.append(smallpatch)
            zout.append(smallzout)
            rowout.append(smallrowout)
            column.append(smallcolumn)

        else:
            patchx = sliceregion.shape[3] // self.n_tiles[2]
            patchy = sliceregion.shape[2] // self.n_tiles[1]
            patchz = sliceregion.shape[1] // self.n_tiles[0]

            if patchx > self.imagex and patchy > self.imagey and patchz > self.imagez:
                if self.overlap_percent > 1 or self.overlap_percent < 0:
                    self.overlap_percent = 0.8

                jumpx = int(self.overlap_percent * patchx)
                jumpy = int(self.overlap_percent * patchy)
                jumpz = int(self.overlap_percent * patchz)
                patchshape = (patchz, patchy, patchx)
                rowstart = 0
                colstart = 0
                zstart = 0
                pairs = []
                # row is y, col is x
                while zstart < sliceregion.shape[1]:
                    rowstart = 0
                    while rowstart < sliceregion.shape[2]:
                        colstart = 0
                        while colstart < sliceregion.shape[3]:
                            # Start iterating over the tile with jumps = stride of the fully convolutional network.
                            pairs.append([zstart, rowstart, colstart])
                            colstart += jumpx
                        rowstart += jumpy
                    zstart +=jumpz
                    # Include the last patch
                while zstart < sliceregion.shape[1]:    
                    rowstart = sliceregion.shape[2] - patchy
                    colstart = 0
                    while colstart < sliceregion.shape[3] - patchx:
                        pairs.append([zstart, rowstart, colstart])
                        colstart += jumpx
                    zstart +=jumpz 
                while zstart < sliceregion.shape[1]:
                    rowstart = 0
                    colstart = sliceregion.shape[2] - patchx
                    while rowstart < sliceregion.shape[1] - patchy:
                        pairs.append([zstart, rowstart, colstart])
                        rowstart += jumpy
                    zstart +=jumpz    

                if sliceregion.shape[1] >= self.imagez and sliceregion.shape[2] >= self.imagey and sliceregion.shape[3] >= self.imagex:

                    patch = []
                    zout = []
                    rowout = []
                    column = []
                    for pair in pairs:
                        smallpatch, smallzout, smallrowout, smallcolumn = chunk_list(sliceregion, patchshape, pair)
                        if smallpatch.shape[1] >= self.imagez and smallpatch.shape[2] >= self.imagey and smallpatch.shape[3] >= self.imagex:
                            patch.append(smallpatch)
                            zout.append(smallzout)
                            rowout.append(smallrowout)
                            column.append(smallcolumn)

            else:

                patch = []
                rowout = []
                zout = []
                column = []
                patchx = sliceregion.shape[3] // self.n_tiles[2]
                patchy = sliceregion.shape[2] // self.n_tiles[1]
                patchz = sliceregion.shape[1] // self.n_tiles[0]
                patchshape = (patchz, patchy, patchx)
                smallpatch, smallzout,  smallrowout, smallcolumn = chunk_list(sliceregion, patchshape, [0, 0, 0])
                patch.append(smallpatch)
                zout.append(smallzout)
                rowout.append(smallrowout)
                column.append(smallcolumn)
        self.patch = patch
        self.sz = zout
        self.sy = rowout
        self.sx = column

    def predict_main(self, sliceregion):
        try:
            self.overlaptiles(sliceregion)
            predictions = []
            allx = []
            ally = []
            allz = []
            if len(self.patch) > 0:
                for i in range(0, len(self.patch)):
                    sum_time_prediction = self.make_patches(self.patch[i])
                    predictions.append(sum_time_prediction)
                    allx.append(self.sx[i])
                    ally.append(self.sy[i])
                    allz.append(self.sz[i])


            else:

                sum_time_prediction = self.make_patches(self.patch)
                predictions.append(sum_time_prediction)
                allx.append(self.sx)
                ally.append(self.sy)
                allz.append(self.sz)

        except tf.errors.ResourceExhaustedError:

            print('Out of memory, increasing overlapping tiles for prediction')
            self.list_n_tiles = list(self.n_tiles)
            self.list_n_tiles[0] = self.n_tiles[0] + 1
            self.list_n_tiles[1] = self.n_tiles[1] + 1
            self.list_n_tiles[2] = self.n_tiles[2] + 1
            self.n_tiles = tuple(self.list_n_tiles)

            self.predict_main(sliceregion)

        return predictions, allx, ally, allz

    def make_patches(self, sliceregion):

        smallimg = np.expand_dims(sliceregion, 0)
        smallimg = tf.reshape(smallimg, (smallimg.shape[0], smallimg.shape[2], smallimg.shape[3],smallimg.shape[4], smallimg.shape[1]))
        prediction_vector = self.model.predict(smallimg, verbose=0)

        return prediction_vector

   


def CreateVolume(patch, size_tminus, size_tplus, timepoint):
    starttime = timepoint - int(size_tminus)
    endtime = timepoint + int(size_tplus) + 1
    #TZYX needs to be reshaed to ZYXT
    smallimg = patch[starttime:endtime, :]
    return smallimg


def chunk_list(image, patchshape, pair):

    zstart = pair[0] 
    rowstart = pair[1]
    colstart = pair[2]

    endz = zstart + patchshape[0]
    endrow = rowstart + patchshape[1]
    endcol = colstart + patchshape[2]

    if endrow > image.shape[2]:
        endrow = image.shape[2]
    if endcol > image.shape[3]:
        endcol = image.shape[3]
    if endz > image.shape[1]:
         endz = image.shape[1]    

    region = (slice(0, image.shape[0]), slice(zstart, endz), slice(rowstart, endrow),
              slice(colstart, endcol))

    # The actual pixels in that region.
    patch = image[region]

    # Always normalize patch that goes into the netowrk for getting a prediction score

    return patch, zstart, rowstart, colstart